"""Model data writer module."""

import json
import logging
from pathlib import Path

import astropy.units as u
import numpy as np
from astropy.io.registry.base import IORegistryError

import simtools.utils.general as gen
from simtools.data_model import schema, validate_data
from simtools.data_model.metadata_collector import MetadataCollector
from simtools.db import db_handler
from simtools.io_operations import io_handler
from simtools.utils import names, value_conversion

__all__ = ["ModelDataWriter"]


class JsonNumpyEncoder(json.JSONEncoder):
    """Convert numpy to python types as accepted by json.dump."""

    def default(self, o):
        if isinstance(o, np.floating):
            return float(o)
        if isinstance(o, np.integer):
            return int(o)
        if isinstance(o, np.ndarray):
            return o.tolist()
        if isinstance(o, u.core.CompositeUnit | u.core.IrreducibleUnit | u.core.Unit):
            return str(o) if o != u.dimensionless_unscaled else None
        if np.issubdtype(type(o), np.bool_):
            return bool(o)
        return super().default(o)


class ModelDataWriter:
    """
    Writer for simulation model data and metadata.

    Parameters
    ----------
    product_data_file: str
        Name of output file.
    product_data_format: str
        Format of output file.
    args_dict: Dictionary
        Dictionary with configuration parameters.
    output_path: str or Path
        Path to output file.
    use_plain_output_path: bool
        Use plain output path.
    args_dict: dict
        Dictionary with configuration parameters.

    """

    def __init__(
        self,
        product_data_file=None,
        product_data_format=None,
        output_path=None,
        use_plain_output_path=True,
        args_dict=None,
    ):
        """Initialize model data writer."""
        self._logger = logging.getLogger(__name__)
        self.io_handler = io_handler.IOHandler()
        self.schema_dict = {}
        if args_dict is not None:
            output_path = args_dict.get("output_path", output_path)
            use_plain_output_path = args_dict.get("use_plain_output_path", use_plain_output_path)
        if output_path is not None:
            self.io_handler.set_paths(
                output_path=output_path, use_plain_output_path=use_plain_output_path
            )
        try:
            self.product_data_file = self.io_handler.get_output_file(file_name=product_data_file)
        except TypeError:
            self.product_data_file = None
        self.product_data_format = self._astropy_data_format(product_data_format)

    @staticmethod
    def dump(
        args_dict, output_file=None, metadata=None, product_data=None, validate_schema_file=None
    ):
        """
        Write model data and metadata (as static method).

        Parameters
        ----------
        args_dict: dict
            Dictionary with configuration parameters (including output file name and path).
        output_file: string or Path
            Name of output file (args["output_file"] is used if this parameter is not set).
        metadata: MetadataCollector object
            Metadata to be written.
        product_data: astropy Table
            Model data to be written
        validate_schema_file: str
            Schema file used in validation of output data.

        """
        writer = ModelDataWriter(
            product_data_file=(
                args_dict.get("output_file", None) if output_file is None else output_file
            ),
            product_data_format=args_dict.get("output_file_format", "ascii.ecsv"),
            args_dict=args_dict,
        )
        if validate_schema_file is not None and not args_dict.get("skip_output_validation", True):
            product_data = writer.validate_and_transform(
                product_data_table=product_data,
                validate_schema_file=validate_schema_file,
            )
        writer.write(metadata=metadata, product_data=product_data)

    @staticmethod
    def dump_model_parameter(
        parameter_name,
        value,
        instrument,
        parameter_version,
        output_file,
        output_path=None,
        use_plain_output_path=False,
        metadata_input_dict=None,
        db_config=None,
    ):
        """
        Generate DB-style model parameter dict and write it to json file.

        Parameters
        ----------
        parameter_name: str
            Name of the parameter.
        value: any
            Value of the parameter.
        instrument: str
            Name of the instrument.
        parameter_version: str
            Version of the parameter.
        output_file: str
            Name of output file.
        output_path: str or Path
            Path to output file.
        use_plain_output_path: bool
            Use plain output path.
        metadata_input_dict: dict
            Input to metadata collector.
        db_config: dict
            Database configuration. If not None, check if parameter with the same version exists.

        Returns
        -------
        dict
            Validated parameter dictionary.
        """
        writer = ModelDataWriter(
            product_data_file=output_file,
            product_data_format="json",
            args_dict=None,
            output_path=output_path,
            use_plain_output_path=use_plain_output_path,
        )
        if db_config is not None:
            writer.check_db_for_existing_parameter(
                parameter_name, instrument, parameter_version, db_config
            )

        unique_id = None
        if metadata_input_dict is not None:
            metadata_input_dict["output_file"] = output_file
            metadata_input_dict["output_file_format"] = Path(output_file).suffix.lstrip(".")
            metadata = MetadataCollector(args_dict=metadata_input_dict)
            metadata.write(output_path / Path(output_file))
            unique_id = (
                metadata.get_top_level_metadata().get("cta", {}).get("product", {}).get("id")
            )

        _json_dict = writer.get_validated_parameter_dict(
            parameter_name, value, instrument, parameter_version, unique_id
        )
        writer.write_dict_to_model_parameter_json(output_file, _json_dict)
        return _json_dict

    def check_db_for_existing_parameter(
        self, parameter_name, instrument, parameter_version, db_config
    ):
        """
        Check if a parameter with the same version exists in the simulation model database.

        Parameters
        ----------
        parameter_name: str
            Name of the parameter.
        instrument: str
            Name of the instrument.
        parameter_version: str
            Version of the parameter.
        db_config: dict
            Database configuration.

        Raises
        ------
        ValueError
            If parameter with the same version exists in the database.
        """
        db = db_handler.DatabaseHandler(mongo_db_config=db_config)
        try:
            db.get_model_parameter(
                parameter=parameter_name,
                parameter_version=parameter_version,
                site=names.get_site_from_array_element_name(instrument),
                array_element_name=instrument,
            )
        except ValueError:
            pass  # parameter does not exist - expected behavior
        else:
            raise ValueError(
                f"Parameter {parameter_name} with version {parameter_version} already exists."
            )

    def get_validated_parameter_dict(
        self,
        parameter_name,
        value,
        instrument,
        parameter_version,
        unique_id=None,
        schema_version=None,
    ):
        """
        Get validated parameter dictionary.

        Parameters
        ----------
        parameter_name: str
            Name of the parameter.
        value: any
            Value of the parameter.
        instrument: str
            Name of the instrument.
        parameter_version: str
            Version of the parameter.
        schema_version: str
            Version of the schema.

        Returns
        -------
        dict
            Validated parameter dictionary.
        """
        self._logger.debug(f"Getting validated parameter dictionary for {instrument}")
        schema_file = schema.get_model_parameter_schema_file(parameter_name)
        self.schema_dict = gen.collect_data_from_file(schema_file)

        value, unit = value_conversion.split_value_and_unit(value)

        data_dict = {
            "schema_version": schema.get_model_parameter_schema_version(schema_version),
            "parameter": parameter_name,
            "instrument": instrument,
            "site": names.get_site_from_array_element_name(instrument),
            "parameter_version": parameter_version,
            "unique_id": unique_id,
            "value": value,
            "unit": unit,
            "type": self._get_parameter_type(),
            "file": self._parameter_is_a_file(),
        }
        return self.validate_and_transform(
            product_data_dict=data_dict,
            validate_schema_file=schema_file,
            is_model_parameter=True,
        )

    def _get_parameter_type(self):
        """
        Return parameter type from schema.

        Returns
        -------
        str
            Parameter type
        """
        _parameter_type = []
        for data in self.schema_dict["data"]:
            _parameter_type.append(data["type"])
        return _parameter_type if len(_parameter_type) > 1 else _parameter_type[0]

    def _parameter_is_a_file(self):
        """
        Check if parameter is a file.

        Returns
        -------
        bool
            True if parameter is a file.

        """
        try:
            return self.schema_dict["data"][0]["type"] == "file"
        except (KeyError, IndexError):
            pass
        return False

    def _get_unit_from_schema(self):
        """
        Return unit(s) from schema dict.

        Returns
        -------
        str or list
            Parameter unit(s)
        """
        try:
            unit_list = []
            for data in self.schema_dict["data"]:
                unit_list.append(data["unit"] if data["unit"] != "dimensionless" else None)
            return unit_list if len(unit_list) > 1 else unit_list[0]
        except (KeyError, IndexError):
            pass
        return None

    def validate_and_transform(
        self,
        product_data_table=None,
        product_data_dict=None,
        validate_schema_file=None,
        is_model_parameter=False,
    ):
        """
        Validate product data using jsonschema given in metadata.

        If necessary, transform product data to match schema.

        Parameters
        ----------
        product_data_table: astropy Table
            Model data to be validated.
        product_data_dict: dict
            Model data to be validated.
        validate_schema_file: str
            Schema file used in validation of output data.
        is_model_parameter: bool
            True if data describes a model parameter.
        """
        _validator = validate_data.DataValidator(
            schema_file=validate_schema_file,
            data_table=product_data_table,
            data_dict=product_data_dict,
            check_exact_data_type=False,
        )
        return _validator.validate_and_transform(is_model_parameter)

    def write(self, product_data=None, metadata=None):
        """
        Write model data and metadata.

        Parameters
        ----------
        product_data: astropy Table
            Model data to be written
        metadata: MetadataCollector object
            Metadata to be written.

        Raises
        ------
        FileNotFoundError
            if data writing was not successful.

        """
        if product_data is None:
            return

        if metadata is not None:
            product_data.meta.update(
                gen.change_dict_keys_case(metadata.get_top_level_metadata(), False)
            )

        self._logger.info(f"Writing data to {self.product_data_file}")
        if isinstance(product_data, dict) and Path(self.product_data_file).suffix == ".json":
            self.write_dict_to_model_parameter_json(self.product_data_file, product_data)
            return
        try:
            product_data.write(
                self.product_data_file, format=self.product_data_format, overwrite=True
            )
        except IORegistryError:
            self._logger.error(f"Error writing model data to {self.product_data_file}.")
            raise
        if metadata is not None:
            metadata.write(self.product_data_file, add_activity_name=True)

    def write_dict_to_model_parameter_json(self, file_name, data_dict):
        """
        Write dictionary to model-parameter-style json file.

        Parameters
        ----------
        file_name : str
            Name of output file.
        data_dict : dict
            Data dictionary.

        Raises
        ------
        FileNotFoundError
            if data writing was not successful.
        """
        data_dict = ModelDataWriter.prepare_data_dict_for_writing(data_dict)
        try:
            self._logger.info(f"Writing data to {self.io_handler.get_output_file(file_name)}")
            with open(self.io_handler.get_output_file(file_name), "w", encoding="UTF-8") as file:
                json.dump(data_dict, file, indent=4, sort_keys=False, cls=JsonNumpyEncoder)
                file.write("\n")
        except FileNotFoundError as exc:
            raise FileNotFoundError(
                f"Error writing model data to {self.io_handler.get_output_file(file_name)}"
            ) from exc

    @staticmethod
    def prepare_data_dict_for_writing(data_dict):
        """
        Prepare data dictionary for writing to json file.

        Ensure sim_telarray style lists as strings.
        Replace "None" with "null" for unit field.

        Parameters
        ----------
        data_dict: dict
            Dictionary with lists.

        Returns
        -------
        dict
            Dictionary with lists converted to strings.

        """
        try:
            data_dict["value"] = gen.convert_list_to_string(data_dict["value"])
            data_dict["unit"] = gen.convert_list_to_string(data_dict["unit"], comma_separated=True)
            data_dict["type"] = gen.convert_list_to_string(
                data_dict["type"], comma_separated=True, collapse_list=True
            )
            if isinstance(data_dict["unit"], str):
                data_dict["unit"] = data_dict["unit"].replace("None", "null")
        except KeyError:
            pass
        return data_dict

    @staticmethod
    def _astropy_data_format(product_data_format):
        """
        Ensure conformance with astropy data format naming.

        Parameters
        ----------
        product_data_format: string
            format identifier

        """
        if product_data_format == "ecsv":
            product_data_format = "ascii.ecsv"
        return product_data_format



#!/usr/bin/python3

r"""
Produce a model parameter report per array element.

The markdown reports include detailed information on each parameter,
comparing their values over various model versions.
Currently only implemented for telescopes.
"""

import logging

from simtools.configuration import configurator
from simtools.io_operations import io_handler
from simtools.reporting.docs_read_parameters import ReadParameters
from simtools.utils import general as gen


def _parse(label):
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=label,
        description=("Produce a markdown report for model parameters."),
    )

    return config.initialize(db_config=True, simulation_model=["site", "telescope"])


def main():  # noqa: D103
    label_name = "reports"
    args, db_config = _parse(label_name)
    io_handler_instance = io_handler.IOHandler()
    output_path = io_handler_instance.get_output_directory(label=label_name, sub_dir="parameters")

    logger = logging.getLogger()
    logger.setLevel(gen.get_log_level_from_user(args["log_level"]))

    ReadParameters(
        db_config,
        args,
        output_path,
    ).produce_model_parameter_reports()

    logger.info(
        f"Markdown report generated for {args['site']} Telescope {args['telescope']}: {output_path}"
    )


if __name__ == "__main__":
    main()



#!/usr/bin/python3
"""Base class for simulation model parameters."""

import logging
import shutil
from copy import copy

import astropy.units as u

import simtools.utils.general as gen
from simtools.db import db_handler
from simtools.io_operations import io_handler
from simtools.simtel.simtel_config_writer import SimtelConfigWriter
from simtools.utils import names

__all__ = ["InvalidModelParameterError", "ModelParameter"]


class InvalidModelParameterError(Exception):
    """Exception for invalid model parameter."""


class ModelParameter:
    """
    Base class for simulation model parameters.

    Provides methods to read and manipulate parameters from DB.

    Parameters
    ----------
    db: DatabaseHandler
        Database handler.
    model_version: str
        Version of the model (ex. 5.0.0).
    site: str
        Site name (e.g., South or North).
    array_element_name: str
        Array element name (e.g., LSTN-01, LSTN-design, ILLN-01).
    collection: str
        instrument class (e.g. telescopes, calibration_devices)
        as stored under collection in the DB.
    mongo_db_config: dict
        MongoDB configuration.
    label: str
        Instance label. Important for output file naming.

    """

    def __init__(
        self,
        mongo_db_config,
        model_version,
        site=None,
        array_element_name=None,
        collection="telescopes",
        db=None,
        label=None,
    ):
        self._logger = logging.getLogger(__name__)
        self._extra_label = None
        self.io_handler = io_handler.IOHandler()
        self.db = (
            db if db is not None else db_handler.DatabaseHandler(mongo_db_config=mongo_db_config)
        )

        self._parameters = {}
        self._simulation_config_parameters = {"corsika": {}, "simtel": {}}
        self.collection = collection
        self.label = label
        self.model_version = model_version
        self.site = names.validate_site_name(site) if site is not None else None
        self.name = (
            names.validate_array_element_name(array_element_name)
            if array_element_name is not None
            else None
        )
        self.design_model = self.db.get_design_model(
            self.model_version, self.name, collection="telescopes"
        )
        self._config_file_directory = None
        self._config_file_path = None
        self._load_parameters_from_db()

        self.simtel_config_writer = None
        self._added_parameter_files = None
        self._is_config_file_up_to_date = False
        self._is_exported_model_files_up_to_date = False

    def _get_parameter_dict(self, par_name):
        """
        Get model parameter dictionary as stored in the DB.

        No conversion to values are applied for the use in simtools
        (e.g., no conversion from the string representation of lists
        to lists).

        Parameters
        ----------
        par_name: str
            Name of the parameter.

        Returns
        -------
        dict
            Dictionary with complete DB entry for the given parameter.

        Raises
        ------
        InvalidModelParameterError
            If par_name does not match any parameter in this model.
        """
        try:
            return self._parameters[par_name]
        except (KeyError, ValueError) as e:
            msg = f"Parameter {par_name} was not found in the model {self.name}, {self.site}."
            self._logger.error(msg)
            raise InvalidModelParameterError(msg) from e

    def get_parameter_value(self, par_name, parameter_dict=None):
        """
        Get the value of a model parameter.

        List of values stored in strings are returns as lists, so that no knowledge
        of the database structure is needed when accessing the model parameters.

        Parameters
        ----------
        par_name: str
            Name of the parameter.
        parameter_dict: dict
            Dictionary with complete DB entry for the given parameter
            (including the 'value', 'units' fields).

        Returns
        -------
        Value of the parameter.

        Raises
        ------
        KeyError
            If par_name does not match any parameter in this model.
        """
        parameter_dict = parameter_dict if parameter_dict else self._get_parameter_dict(par_name)
        try:
            _parameter = parameter_dict["value"]
        except KeyError as exc:
            self._logger.error(f"Parameter {par_name} does not have a value")
            raise exc
        if isinstance(_parameter, str):
            _is_float = False
            try:
                _is_float = self.get_parameter_type(par_name).startswith("float")
            except (InvalidModelParameterError, TypeError):  # float - in case we don't know
                _is_float = True
            _parameter = gen.convert_string_to_list(_parameter, is_float=_is_float)
            _parameter = _parameter if len(_parameter) > 1 else _parameter[0]

        return _parameter

    def get_parameter_value_with_unit(self, par_name):
        """
        Get the value of an existing parameter of the model as an Astropy Quantity with its unit.

        Parameters
        ----------
        par_name: str
            Name of the parameter.

        Returns
        -------
        Astropy quantity with the value of the parameter multiplied by its unit.
        If no unit is provided in the model, the value is returned without a unit.

        """
        _parameter = self._get_parameter_dict(par_name)
        _value = self.get_parameter_value(par_name, _parameter)

        try:
            if isinstance(_parameter.get("unit"), str):
                _unit = [item.strip() for item in _parameter.get("unit").split(",")]
            else:
                _unit = _parameter.get("unit")

            # if there is only one value or the values share one unit
            if (isinstance(_value, (int | float))) or (len(_value) > len(_unit)):
                return _value * u.Unit(_unit[0])

            # entries with 'null' units should be returned as dimensionless
            _astropy_units = [
                u.Unit(item) if item != "null" else u.dimensionless_unscaled for item in _unit
            ]

            return [_value[i] * _astropy_units[i] for i in range(len(_value))]

        except (KeyError, TypeError, AttributeError) as exc:
            self._logger.debug(f"{exc} encountered, returning only value without units.")
            return _value  # if unit is NoneType

    def get_parameter_type(self, par_name):
        """
        Get the type of existing parameter of the model (value of 'type' field of DB entry).

        Parameters
        ----------
        par_name: str
            Name of the parameter.

        Returns
        -------
        str or None
            type of the parameter (None if no type is defined)

        """
        parameter_dict = self._get_parameter_dict(par_name)
        try:
            return parameter_dict["type"]
        except KeyError:
            self._logger.debug(f"Parameter {par_name} does not have a type.")
        return None

    def get_parameter_file_flag(self, par_name):
        """
        Get value of parameter file flag of this database entry (boolean 'file' field of DB entry).

        Parameters
        ----------
        par_name: str
            Name of the parameter.

        Returns
        -------
        bool
            True if file flag is set.

        """
        parameter_dict = self._get_parameter_dict(par_name)
        try:
            return parameter_dict["file"]
        except KeyError:
            self._logger.debug(f"Parameter {par_name} does not have a file associated with it.")
        return False

    def get_parameter_version(self, par_name):
        """
        Get version for a given parameter used in the model.

        Parameters
        ----------
        par_name: str
            Name of the parameter.

        Returns
        -------
        str
            parameter version used in the model (eg. '1.0.0')
        """
        return self._get_parameter_dict(par_name)["parameter_version"]

    def print_parameters(self):
        """Print parameters and their values for debugging purposes."""
        for par in self._parameters:
            print(f"{par} = {self.get_parameter_value(par)}")

    def _set_config_file_directory_and_name(self):
        """Set and create the directory and the name of the config file."""
        if self.name is None and self.site is None:
            return

        self._config_file_directory = self.io_handler.get_output_directory(
            label=self.label, sub_dir="model"
        )

        # Setting file name and the location
        config_file_name = names.simtel_config_file_name(
            self.site,
            self.model_version,
            telescope_model_name=self.name,
            label=self.label,
            extra_label=self._extra_label,
        )
        self._config_file_path = self.config_file_directory.joinpath(config_file_name)

        self._logger.debug(f"Config file path: {self._config_file_path}")

    def get_simulation_software_parameters(self, simulation_software):
        """
        Get simulation software parameters.

        Parameters
        ----------
        simulation_software: str
            Simulation software name.

        Returns
        -------
        dict
            Simulation software parameters.
        """
        return self._simulation_config_parameters.get(simulation_software)

    def has_parameter(self, par_name):
        """Check if a parameter exists in the model.

        Parameters
        ----------
        par_name : str
            Name of the parameter.

        Returns
        -------
        bool
            True if parameter exists in the model.
        """
        return par_name in self._parameters

    def _load_simulation_software_parameter(self):
        """Read simulation software parameters from DB."""
        for simulation_software in self._simulation_config_parameters:
            try:
                self._simulation_config_parameters[simulation_software] = (
                    self.db.get_simulation_configuration_parameters(
                        site=self.site,
                        array_element_name=self.name,
                        model_version=self.model_version,
                        simulation_software=simulation_software,
                    )
                )
            except ValueError as exc:
                self._logger.warning(
                    f"No {simulation_software} parameters found for "
                    f"{self.site}, {self.name} (model version {self.model_version}). "
                    f" (Query {exc})"
                )

    def _load_parameters_from_db(self):
        """Read parameters from DB and store them in _parameters."""
        if self.db is None:
            return

        if self.name is not None:
            self._parameters = self.db.get_model_parameters(
                self.site, self.name, self.collection, self.model_version
            )

        if self.site is not None:
            self._parameters.update(
                self.db.get_model_parameters(
                    self.site,
                    None,
                    "sites",
                    self.model_version,
                )
            )
        self._load_simulation_software_parameter()

    def set_extra_label(self, extra_label):
        """
        Set an extra label for the name of the config file.

        Notes
        -----
        The config file directory name is not affected by the extra label. Only the file name is \
        changed. This is important for the ArrayModel class to export multiple config files in the\
        same directory.

        Parameters
        ----------
        extra_label: str
            Extra label to be appended to the original label.
        """
        self._extra_label = extra_label
        self._set_config_file_directory_and_name()

    @property
    def extra_label(self):
        """Return the extra label if defined, if not return ''."""
        return self._extra_label if self._extra_label is not None else ""

    def get_simtel_parameters(self, parameters=None):
        """
        Get simtel parameters as name and value pairs.

        Parameters
        ----------
        parameters: dict
            Parameters (simtools) to be renamed (if necessary)

        Returns
        -------
        dict
            simtel parameters as dict (sorted by parameter names)

        """
        if parameters is None:
            parameters = self._parameters

        _simtel_parameter_value = {}
        for key in parameters:
            _par_name = names.get_simulation_software_name_from_parameter_name(
                key, simulation_software="sim_telarray"
            )
            if _par_name is not None:
                _simtel_parameter_value[_par_name] = parameters[key].get("value")
        return dict(sorted(_simtel_parameter_value.items()))

    def change_parameter(self, par_name, value):
        """
        Change the value of an existing parameter.

        This function does not modify the  DB, it affects only the current instance.

        Parameters
        ----------
        par_name: str
            Name of the parameter.
        value:
            Value of the parameter.

        Raises
        ------
        InvalidModelParameterError
            If the parameter to be changed does not exist in this model.
        """
        if par_name not in self._parameters:
            msg = f"Parameter {par_name} not in the model"
            self._logger.error(msg)
            raise InvalidModelParameterError(msg)

        if isinstance(value, str):
            value = gen.convert_string_to_list(value)

        if not gen.validate_data_type(
            reference_dtype=self.get_parameter_type(par_name),
            value=value,
            dtype=None,
            allow_subtypes=True,
        ):
            raise ValueError(
                f"Could not cast {value} of type {type(value)} "
                f"to {self.get_parameter_type(par_name)}."
            )

        self._logger.debug(
            f"Changing parameter {par_name} from {self.get_parameter_value(par_name)} to {value}"
        )
        self._parameters[par_name]["value"] = value

        # In case parameter is a file, the model files will be outdated
        if self.get_parameter_file_flag(par_name):
            self._is_exported_model_files_up_to_date = False

        self._is_config_file_up_to_date = False

    def change_multiple_parameters_from_file(self, file_name):
        """
        Change values of multiple existing parameters in the model from a file.

        This function does not modify the DB, it affects only the current instance.
        This feature is intended for developers and lacks validation.

        Parameters
        ----------
        file_name: str
            File containing the parameters to be changed.
        """
        self._logger.warning(
            "Changing multiple parameters from file is a feature for developers."
            "Insufficient validation of parameters."
        )
        self._logger.debug(f"Changing parameters from file {file_name}")
        self.change_multiple_parameters(**gen.collect_data_from_file(file_name=file_name))

    def change_multiple_parameters(self, **kwargs):
        """
        Change the value of multiple existing parameters in the model.

        This function does not modify the DB, it affects only the current instance.

        Parameters
        ----------
        **kwargs
            Parameters should be passed as parameter_name=value.

        """
        for par, value in kwargs.items():
            if par in self._parameters:
                self.change_parameter(par, value)

        self._is_config_file_up_to_date = False

    def export_parameter_file(self, par_name, file_path):
        """
        Export a file to the config file directory.

        Parameters
        ----------
        par_name: str
            Name of the parameter.
        file_path: str
            Path of the file to be added to the config file directory.
        """
        if self._added_parameter_files is None:
            self._added_parameter_files = []
        self._added_parameter_files.append(par_name)
        shutil.copy(file_path, self.config_file_directory)

    def export_model_files(self):
        """Export the model files into the config file directory."""
        # Removing parameter files added manually (which are not in DB)
        pars_from_db = copy(self._parameters)
        if self._added_parameter_files is not None:
            for par in self._added_parameter_files:
                pars_from_db.pop(par)

        self.db.export_model_files(parameters=pars_from_db, dest=self.config_file_directory)
        self._is_exported_model_files_up_to_date = True

    def export_config_file(self):
        """Export the config file used by sim_telarray."""
        # Exporting model file
        if not self._is_exported_model_files_up_to_date:
            self.export_model_files()

        # Using SimtelConfigWriter to write the config file.
        self._load_simtel_config_writer()
        self.simtel_config_writer.write_telescope_config_file(
            config_file_path=self.config_file_path,
            parameters=self.get_simtel_parameters(parameters=self._parameters),
            config_parameters=self.get_simtel_parameters(
                parameters=self._simulation_config_parameters["simtel"]
            ),
        )

    @property
    def config_file_directory(self):
        """Directory for configure files. Configure, if necessary."""
        if self._config_file_directory is None:
            self._set_config_file_directory_and_name()
        return self._config_file_directory

    @property
    def config_file_path(self):
        """Path of the config file. Configure, if necessary."""
        if self._config_file_path is None:
            self._set_config_file_directory_and_name()
        return self._config_file_path

    def get_config_file(self, no_export=False):
        """
        Get the path of the config file for sim_telarray.

        The config file is produced if the file is not up to date.

        Parameters
        ----------
        no_export: bool
            Turn it on if you do not want the file to be exported.

        Returns
        -------
        Path
            Path of the exported config file for sim_telarray.
        """
        if not self._is_config_file_up_to_date and not no_export:
            self.export_config_file()
        return self.config_file_path

    def _load_simtel_config_writer(self):
        """Load the SimtelConfigWriter object."""
        if self.simtel_config_writer is None:
            self.simtel_config_writer = SimtelConfigWriter(
                site=self.site,
                telescope_model_name=self.name,
                model_version=self.model_version,
                label=self.label,
            )

    def export_nsb_spectrum_to_telescope_altitude_correction_file(self, model_directory):
        """
        Export the NSB spectrum to the telescope altitude correction file.

        This method is needed because testeff corrects the NSB spectrum from the original altitude
        used in the Benn & Ellison model to the telescope altitude.
        This is done internally in testeff, but the NSB spectrum is not written out to the model
        directory. This method allows to export it explicitly.

        Parameters
        ----------
        model_directory: Path
            Model directory to export the file to.
        """
        self.db.export_model_files(
            parameters={
                "nsb_spectrum_at_2200m": {
                    "value": self._simulation_config_parameters["simtel"][
                        "correct_nsb_spectrum_to_telescope_altitude"
                    ]["value"],
                    "file": True,
                }
            },
            dest=model_directory,
        )



"""
Metadata collector for simtools.

This should be the only module in simtools with knowledge on the
implementation of the observatory metadata model.

"""

import getpass
import logging
import uuid
from pathlib import Path

import yaml

import simtools.constants
import simtools.utils.general as gen
import simtools.version
from simtools.constants import METADATA_JSON_SCHEMA
from simtools.data_model import metadata_model, schema
from simtools.io_operations import io_handler
from simtools.utils import names

__all__ = ["MetadataCollector"]


class MetadataCollector:
    """
    Collects metadata to describe the current simtools activity and its data products.

    Collect metadata from command line configuration, input data, environment,
    and schema descriptions. Depends on the CTAO top-level metadata definition.

    Two dictionaries store two different types of metadata:

    - top_level_meta: metadata for the current activity
    - input_metadata: metadata from input data

    Parameters
    ----------
    args_dict: dict
        Command line parameters
    metadata_file_name: str
        Name of metadata file (only required when args_dict is None)
    data_model_name: str
        Name of data model parameter
    observatory: str
        Name of observatory (default: "cta")
    clean_meta: bool
        Clean metadata from None values and empty lists (default: True)
    """

    def __init__(
        self,
        args_dict,
        metadata_file_name=None,
        data_model_name=None,
        observatory="cta",
        clean_meta=True,
    ):
        """Initialize metadata collector."""
        self._logger = logging.getLogger(__name__)
        self.observatory = observatory
        self.io_handler = io_handler.IOHandler()

        self.args_dict = args_dict if args_dict else {}
        self.data_model_name = data_model_name
        self.schema_file = None
        self.schema_dict = None
        self.top_level_meta = gen.change_dict_keys_case(
            data_dict=metadata_model.get_default_metadata_dict(), lower_case=True
        )
        self.input_metadata = self._read_input_metadata_from_file(metadata_file_name)
        self.collect_meta_data()
        if clean_meta:
            self.top_level_meta = self.clean_meta_data(self.top_level_meta)

    def collect_meta_data(self):
        """Collect and verify product metadata for each main-level metadata type."""
        meta_types = self.top_level_meta[self.observatory].keys()
        for meta_type in meta_types:
            try:
                fill_method = getattr(self, f"_fill_{meta_type}_meta")
                fill_method(self.top_level_meta[self.observatory][meta_type])
            except AttributeError:
                self._logger.debug(f"Method _fill_{meta_type}_meta not implemented")

    def get_top_level_metadata(self):
        """
        Return top level metadata dictionary (with updated activity end time).

        Returns
        -------
        dict
            Top level metadata dictionary.

        """
        try:
            self.top_level_meta[self.observatory]["activity"]["end"] = (
                gen.now_date_time_in_isoformat()
            )
        except KeyError:
            pass
        return self.top_level_meta

    @staticmethod
    def dump(args_dict, output_file, add_activity_name=False):
        """
        Write metadata to file (static method).

        Parameters
        ----------
        args_dict: dict
            Command line parameters
        output_file: str or Path
            Name of output file.
        add_activity_name: bool
            Add activity name to file name.
        """
        collector = MetadataCollector(args_dict)
        collector.write(output_file, add_activity_name=add_activity_name)

    def write(self, yml_file=None, keys_lower_case=False, add_activity_name=False):
        """
        Write toplevel metadata to file (yaml file format).

        Parameters
        ----------
        metadata: dict
            Metadata to be stored
        yml_file: str
            Name of output file.
        keys_lower_case: bool
            Write yaml keys in lower case.
        add_activity_name: bool
            Add activity name to file name.

        Returns
        -------
        str
            Name of output file

        Raises
        ------
        FileNotFoundError
            If yml_file not found.
        TypeError
            If yml_file is not defined.
        """
        metadata = self.get_top_level_metadata()
        activity_name = metadata.get("cta", {}).get("activity", {}).get("name", "").rstrip(".")
        suffix = f".{activity_name}.meta.yml" if add_activity_name else ".meta.yml"

        if yml_file is None:
            raise TypeError("No output file for metadata defined")

        try:
            yml_file = names.file_name_with_version(yml_file, suffix)
            with open(yml_file, "w", encoding="UTF-8") as file:
                yaml.safe_dump(
                    gen.change_dict_keys_case(
                        gen.remove_substring_recursively_from_dict(metadata, substring="\n"),
                        keys_lower_case,
                    ),
                    file,
                    sort_keys=False,
                    explicit_start=True,
                )
            self._logger.info(f"Writing metadata to {yml_file}")
            return yml_file
        except FileNotFoundError as exc:
            raise FileNotFoundError(f"Error writing metadata to {yml_file}") from exc

    def get_data_model_schema_file_name(self):
        """
        Return data model schema file name.

        The schema file name is taken (in this order) from the command line,
        from the metadata file, from the data model name, or from the input
        metadata file.

        Returns
        -------
        str
            Name of schema file.

        """
        # from command line
        if self.args_dict.get("schema"):
            self._logger.debug(f"Schema file from command line: {self.args_dict['schema']}")
            return self.args_dict["schema"]

        # from metadata
        try:
            url = self.top_level_meta[self.observatory]["product"]["data"]["model"]["url"]
            if url:
                self._logger.debug(f"Schema file from product metadata: {url}")
                return url
        except KeyError:
            pass

        # from data model name
        if self.data_model_name:
            self._logger.debug(f"Schema file from data model name: {self.data_model_name}")
            return str(schema.get_model_parameter_schema_file(self.data_model_name))

        # from first entry in input metadata (least preferred)
        try:
            url = self.input_metadata[0][self.observatory]["product"]["data"]["model"]["url"]
            self._logger.debug(f"Schema file from input metadata: {url}")
            return url
        except (KeyError, TypeError):
            pass

        self._logger.warning("No schema file found.")
        return None

    def get_data_model_schema_dict(self):
        """
        Return data model schema dictionary.

        Returns
        -------
        dict
            Data model schema dictionary.

        """
        try:
            return gen.collect_data_from_file(file_name=self.schema_file)
        except TypeError:
            self._logger.debug(f"No valid schema file provided ({self.schema_file}).")
        return {}

    def get_site(self, from_input_meta=False):
        """
        Get site entry from metadata. Allow to get from collected or from input metadata.

        Parameters
        ----------
        from_input_meta: bool
            Get site from first entry of input metadata (default: False)

        Returns
        -------
        str
            Site name

        """
        try:
            _site = (
                self.top_level_meta[self.observatory]["instrument"]["site"]
                if not from_input_meta
                else self.input_metadata[0][self.observatory]["instrument"]["site"]
            )
            if _site is not None:
                return names.validate_site_name(_site)
        except (KeyError, TypeError):
            pass
        return None

    def _fill_contact_meta(self, contact_dict):
        """
        Fill contact metadata fields. Get user name from system level if not given.

        Parameters
        ----------
        contact_dict: dict
            Dictionary for contact metadata fields.
        """
        contact_dict["name"] = contact_dict.get("name") or self.args_dict.get("user_name")
        if contact_dict["name"] is None:
            self._logger.warning("No user name provided, take user info from system level.")
            try:
                contact_dict["name"] = getpass.getuser()
            except Exception as exc:  # pylint: disable=broad-except
                contact_dict["name"] = "UNKNOWN_USER"
                self._logger.warning(
                    f"Failed to get user name: {exc}, setting it to {contact_dict['name']} "
                )
        meta_dict = {
            "email": "user_mail",
            "orcid": "user_orcid",
            "organization": "user_organization",
        }
        for key, value in meta_dict.items():
            contact_dict[key] = contact_dict.get(key) or self.args_dict.get(value)

    def _fill_context_meta(self, context_dict):
        """
        Fill context metadata fields with product metadata from input data.

        Parameters
        ----------
        context_dict: dict
            Dictionary for context metadata fields.

        """
        input_metadata = (
            self.input_metadata if isinstance(self.input_metadata, list) else [self.input_metadata]
        )

        for metadata in input_metadata:
            try:  # wide try..except as for some cases we expect that there is no product metadata
                reduced_product_meta = {
                    key: value
                    for key, value in metadata[self.observatory]["product"].items()
                    if key in {"description", "id", "creation_time", "valid", "format", "filename"}
                }
                if metadata[self.observatory].get("activity", {}).get("name"):
                    reduced_product_meta["activity_name"] = metadata[self.observatory][
                        "activity"
                    ].get("name")
                context_dict["associated_data"] = self._fill_context_sim_list(
                    context_dict["associated_data"], reduced_product_meta
                )
            except (KeyError, TypeError):
                self._logger.debug("No input product metadata appended to associated data.")

    def _read_input_metadata_from_file(self, metadata_file_name_expression=None):
        """
        Read and validate input metadata from file.

        In case of an ecsv file including a table, the metadata is read from the table meta data.
        Returns empty dict in case no file is given.

        Parameter
        ---------
        metadata_file_name_expression: str or Path
            Name of metadata file (regular expressions allowed).

        Returns
        -------
        dict
            Metadata dictionary.

        Raises
        ------
        gen.InvalidConfigDataError, FileNotFoundError
            if metadata cannot be read from file.
        KeyError:
            if metadata does not exist

        """
        metadata_file_names = (
            metadata_file_name_expression
            or self.args_dict.get("input_meta")
            or self.args_dict.get("input")
        )

        try:
            metadata_files = gen.resolve_file_patterns(metadata_file_names)
        except ValueError:
            self._logger.debug("No input metadata file defined.")
            return None

        metadata = []
        for metadata_file in metadata_files:
            self._logger.debug(f"Reading meta data from {metadata_file}")
            if Path(metadata_file).suffix in (".yaml", ".yml", ".json"):
                _input_metadata = self._read_input_metadata_from_yml_or_json(metadata_file)
            elif Path(metadata_file).suffix == ".ecsv":
                _input_metadata = self._read_input_metadata_from_ecsv(metadata_file)
            else:
                raise gen.InvalidConfigDataError(f"Unknown metadata file format: {metadata_file}")

            schema.validate_dict_using_schema(_input_metadata, schema_file=METADATA_JSON_SCHEMA)
            metadata.append(gen.change_dict_keys_case(_input_metadata, lower_case=True))

        return metadata

    def _read_input_metadata_from_ecsv(self, metadata_file_name):
        """Read input metadata from ecsv file."""
        from astropy.table import Table  # pylint: disable=C0415

        try:
            return {
                self.observatory.upper(): Table.read(metadata_file_name).meta[
                    self.observatory.upper()
                ]
            }
        except (FileNotFoundError, KeyError, AttributeError) as exc:
            self._logger.error(
                "Failed reading metadata for %s from %s", self.observatory, metadata_file_name
            )
            raise exc

    def _read_input_metadata_from_yml_or_json(self, metadata_file_name):
        """Read input metadata from yml or json file."""
        try:
            _input_metadata = gen.collect_data_from_file(file_name=metadata_file_name)
            _json_type_metadata = {"Metadata", "metadata", "METADATA"}.intersection(_input_metadata)
            if len(_json_type_metadata) == 1:
                _input_metadata = _input_metadata[_json_type_metadata.pop()]
            if len(_json_type_metadata) > 1:
                self._logger.error("More than one metadata entry found in %s", metadata_file_name)
                raise gen.InvalidConfigDataError
        except (gen.InvalidConfigDataError, FileNotFoundError) as exc:
            self._logger.error("Failed reading metadata from %s", metadata_file_name)
            raise exc
        return _input_metadata

    def _fill_product_meta(self, product_dict):
        """
        Fill metadata for data products fields.

        If a schema file is given for the data products, try and read product:data:model metadata
        from there.

        Parameters
        ----------
        product_dict: dict
            Dictionary describing data product.

        Raises
        ------
        KeyError
            if relevant fields are not defined in top level metadata dictionary.

        """
        self.schema_file = self.get_data_model_schema_file_name()
        self.schema_dict = self.get_data_model_schema_dict()

        product_dict["id"] = str(uuid.uuid4())
        product_dict["creation_time"] = gen.now_date_time_in_isoformat()
        product_dict["description"] = self.schema_dict.get("description", None)

        # DATA:CATEGORY
        product_dict["data"]["category"] = "SIM"
        product_dict["data"]["level"] = "R1"
        product_dict["data"]["type"] = "Service"
        try:
            product_dict["data"]["association"] = self.schema_dict["instrument"]["class"]
        except KeyError:
            pass

        # DATA:MODEL
        product_dict["data"]["model"]["name"] = (
            self.schema_dict.get("name")
            or self.args_dict.get("metadata_product_data_name")
            or "undefined_model_name"
        )
        product_dict["data"]["model"]["version"] = self.schema_dict.get("version", "0.0.0")
        product_dict["data"]["model"]["type"] = self.schema_dict.get("meta_schema", None)
        product_dict["data"]["model"]["url"] = self.schema_file or self.args_dict.get(
            "metadata_product_data_url"
        )

        product_dict["filename"] = str(self.args_dict.get("output_file", ""))
        product_dict["format"] = (
            self.args_dict.get("output_file_format")
            or Path(product_dict["filename"]).suffix.lstrip(".")
            or None
        )

    def _fill_instrument_meta(self, instrument_dict):
        """
        Fill instrument metadata fields.

        Note inconsistency in command line arguments for 'ID',
        which is either 'instrument' or 'telescope'.

        Parameters
        ----------
        instrument_dict: dict
            Dictionary for instrument metadata fields.

        """
        instrument_dict["site"] = self.args_dict.get("site")
        instrument_dict["ID"] = self.args_dict.get("instrument") or self.args_dict.get("telescope")
        if instrument_dict["ID"]:
            instrument_dict["class"] = names.get_collection_name_from_array_element_name(
                instrument_dict["ID"], False
            )
            instrument_dict["type"] = (
                names.get_array_element_type_from_name(instrument_dict["ID"])
                if not instrument_dict.get("type")
                else instrument_dict["type"]
            )

    def _fill_process_meta(self, process_dict):
        """
        Fill process fields in metadata.

        Parameters
        ----------
        process_dict: dict
            Dictionary for process metadata fields.

        """
        process_dict["type"] = "simulation"

    def _fill_activity_meta(self, activity_dict):
        """
        Fill activity (software) related metadata.

        Parameters
        ----------
        activity_dict: dict
            Dictionary for top-level activity metadata.

        """
        activity_dict["name"] = self.args_dict.get("label", None)
        activity_dict["type"] = "software"
        activity_dict["id"] = self.args_dict.get("activity_id", "UNDEFINED_ACTIVITY_ID")
        activity_dict["start"] = gen.now_date_time_in_isoformat()
        activity_dict["end"] = activity_dict["start"]
        activity_dict["software"]["name"] = "simtools"
        activity_dict["software"]["version"] = simtools.version.__version__

    def _merge_config_dicts(self, dict_high, dict_low, add_new_fields=False):
        """
        Merge two config dicts and replace values in dict_high which are Nonetype.

        Priority to dict_high in case of conflicting entries.

        Parameters
        ----------
        dict_high: dict
            Dictionary into which values are merged.
        dict_low: dict
            Dictionary from which values are taken for merging.
        add_new_fields: bool
            If true: add fields from dict_low to dict_high, if they don't exist in dict_high

        """
        try:
            for k in dict_low:
                if k in dict_high:
                    if isinstance(dict_low[k], dict):
                        self._merge_config_dicts(dict_high[k], dict_low[k], add_new_fields)
                    elif dict_high[k] is None:
                        dict_high[k] = dict_low[k]
                    elif dict_high[k] != dict_low[k] and dict_low[k] is not None:
                        self._logger.debug(
                            f"Conflicting entries between dict: {dict_high[k]} vs {dict_low[k]} "
                            f"(use {dict_high[k]})"
                        )
                elif add_new_fields:
                    dict_high[k] = dict_low[k]
        except TypeError as exc:
            raise TypeError("Error merging dictionaries") from exc

    def _fill_context_sim_list(self, meta_list, new_entry_dict):
        """
        Fill list-type entries into metadata.

        Take into account the first list entry is the default value filled with Nones.

        Parameters
        ----------
        meta_list: list
            List of metadata entries.
        new_entry_dict: dict
            New metadata entry to be added to meta_list.

        Returns
        -------
        list
            Updated meta list.

        """
        if not new_entry_dict:
            return []
        if meta_list is None or not meta_list:
            return [new_entry_dict]
        if self._all_values_none(meta_list[0]):
            meta_list[0] = new_entry_dict
        else:
            meta_list.append(new_entry_dict)
        return meta_list

    def _process_metadata_from_file(self, meta_dict):
        """
        Process metadata from file to ensure compatibility with metadata model.

        Changes keys to lower case and removes line feeds from description fields.

        Parameters
        ----------
        meta_dict: dict
            Input metadata dictionary.

        Returns
        -------
        dict
            Metadata dictionary.

        """
        meta_dict = gen.change_dict_keys_case(meta_dict, True)
        try:
            meta_dict[self.observatory]["product"]["description"] = self._remove_line_feed(
                meta_dict[self.observatory]["product"]["description"]
            )
        except (KeyError, AttributeError):
            pass

        return meta_dict

    @staticmethod
    def _remove_line_feed(string):
        """
        Remove all line feeds from a string.

        Parameters
        ----------
        str
            input string

        Returns
        -------
        str
            with line feeds removed
        """
        return string.replace("\n", " ").replace("\r", "").replace("  ", " ")

    def _copy_list_type_metadata(self, context_dict, _input_metadata, key):
        """
        Copy list-type metadata from file.

        Very fine tuned.

        Parameters
        ----------
        context_dict: dict
            Dictionary for top level metadata (context level)
        _input_metadata: dict
            Dictionary for metadata from file.
        key: str
            Key for metadata entry.

        """
        try:
            for document in _input_metadata["context"][key]:
                context_dict[key] = self._fill_context_sim_list(context_dict[key], document)
        except KeyError:
            pass

    def _all_values_none(self, input_dict):
        """
        Check recursively if all values in a dictionary are None.

        Parameters
        ----------
        input_dict: dict
            Input dictionary.

        Returns
        -------
        bool
            True if all entries are None.

        """
        if not isinstance(input_dict, dict):
            return input_dict is None

        return all(self._all_values_none(value) for value in input_dict.values())

    def clean_meta_data(self, meta_dict):
        """
        Clean metadata dictionary from None values and empty lists.

        Parameters
        ----------
        meta_dict: dict
            Metadata dictionary.

        """

        def clean_list(value):
            nested_list = [
                self.clean_meta_data(item) if isinstance(item, dict) else item for item in value
            ]
            return [item for item in nested_list if item not in (None, "", [], {})]

        cleaned = {}
        for key, value in meta_dict.items():
            if value in (None, []):
                continue
            if isinstance(value, dict):
                nested = self.clean_meta_data(value)
                if nested:  # Only add if not empty
                    cleaned[key] = nested
            elif isinstance(value, list):
                nested_list = clean_list(value)
                if nested_list:  # Only add if not empty
                    cleaned[key] = nested_list
            else:
                cleaned[key] = value
        return cleaned



"""Simulation runner for ray tracing simulations."""

import logging
from collections import namedtuple

import astropy.units as u

from simtools.io_operations import io_handler
from simtools.runners.simtel_runner import SimtelRunner
from simtools.utils import names
from simtools.utils.general import clear_default_sim_telarray_cfg_directories

__all__ = ["SimulatorRayTracing"]

# pylint: disable=no-member
# The line above is needed because there are members which are created
# by adding them to the __dict__ of the class rather than directly.


class SimulatorRayTracing(SimtelRunner):
    """
    Perform ray tracing simulations with sim_telarray.

    Parameters
    ----------
    telescope_model: TelescopeModel
        telescope model
    label: str
        label used for output file naming.
    simtel_path: str or Path
        Location of sim_telarray installation.
    config_data: namedtuple
        namedtuple containing the configurable parameters as values (expected units in
        brackets): zenith_angle (deg), off_axis_angle (deg), source_distance (km),
        single_mirror_mode, use_random_focal_length,
        mirror_numbers.
    force_simulate: bool
        Remove existing files and force re-running of the ray-tracing simulation.
    """

    def __init__(
        self,
        telescope_model,
        label=None,
        simtel_path=None,
        config_data=None,
        force_simulate=False,
        test=False,
    ):
        """Initialize SimtelRunner."""
        self._logger = logging.getLogger(__name__)
        self._logger.debug("Init SimulatorRayTracing")

        super().__init__(label=label, simtel_path=simtel_path)

        self.telescope_model = telescope_model
        self.label = label if label is not None else self.telescope_model.label

        self.io_handler = io_handler.IOHandler()
        self._base_directory = self.io_handler.get_output_directory(self.label, "ray-tracing")

        self.config = (
            self._config_to_namedtuple(config_data)
            if isinstance(config_data, dict)
            else config_data
        )
        self._rep_number = 0
        self.runs_per_set = 1 if self.config.single_mirror_mode else 20
        self.photons_per_run = 100000 if not test else 5000

        self._load_required_files(force_simulate)

    def _load_required_files(self, force_simulate):
        """
        Load required files for the simulation. Depends on the running mode.

        Initialize files for the simulation.

        Parameters
        ----------
        force_simulate: bool
            Remove existing files and force re-running of the ray-tracing simulation.
        """
        # This file is not actually needed and does not exist in simtools.
        # It is required as CORSIKA input file to sim_telarray
        self._corsika_file = self._simtel_path.joinpath("run9991.corsika.gz")

        # Loop to define and remove existing files.
        # Files will be named _base_file = self.__dict__['_' + base + 'File']
        for base_name in ["stars", "photons", "log"]:
            file_name = names.generate_file_name(
                file_type=base_name,
                suffix=".log" if base_name == "log" else ".lis",
                site=self.telescope_model.site,
                telescope_model_name=self.telescope_model.name,
                source_distance=(
                    None if self.config.single_mirror_mode else self.config.source_distance
                ),
                zenith_angle=self.config.zenith_angle,
                off_axis_angle=self.config.off_axis_angle,
                mirror_number=(
                    self.config.mirror_numbers if self.config.single_mirror_mode else None
                ),
                label=self.label,
            )
            file = self._base_directory.joinpath(file_name)
            if file.exists() and force_simulate:
                file.unlink()
            # Defining the file name variable as an class attribute.
            self.__dict__["_" + base_name + "_file"] = file

        if not file.exists() or force_simulate:
            # Adding header to photon list file.
            with self._photons_file.open("w", encoding="utf-8") as file:
                file.write(f"#{50 * '='}\n")
                file.write("# List of photons for RayTracing simulations\n")
                file.write(f"#{50 * '='}\n")
                file.write(f"# config_file = {self.telescope_model.get_config_file()}\n")
                file.write(f"# zenith_angle [deg] = {self.config.zenith_angle}\n")
                file.write(f"# off_axis_angle [deg] = {self.config.off_axis_angle}\n")
                file.write(f"# source_distance [km] = {self.config.source_distance}\n")
                if self.config.single_mirror_mode:
                    file.write(f"# mirror_number = {self.config.mirror_numbers}\n\n")

            # Filling a star file with a single light source defined by
            # - azimuth
            # - elevation
            # - flux
            # - distance of light source
            with self._stars_file.open("w", encoding="utf-8") as file:
                file.write(
                    f"0. {90.0 - self.config.zenith_angle} 1.0 {self.config.source_distance}\n"
                )

        if self.config.single_mirror_mode:
            self._logger.debug("For single mirror mode, need to prepare the single pixel camera.")
            self._write_out_single_pixel_camera_file()

    def _make_run_command(self, run_number=None, input_file=None):  # pylint: disable=unused-argument
        """Generate simtel_array run command."""
        if self.config.single_mirror_mode:
            # Note: no mirror length defined for dual-mirror telescopes
            _mirror_focal_length = float(
                self.telescope_model.get_parameter_value("mirror_focal_length")
            )

        # RayTracing
        command = str(self._simtel_path.joinpath("sim_telarray/bin/sim_telarray"))
        command += f" -c {self.telescope_model.get_config_file()}"
        command += f" -I{self.telescope_model.config_file_directory}"
        command += super().get_config_option("random_state", "none")
        command += super().get_config_option("IMAGING_LIST", str(self._photons_file))
        command += super().get_config_option("stars", str(self._stars_file))
        command += super().get_config_option(
            "altitude", self.telescope_model.get_parameter_value("corsika_observation_level")
        )
        command += super().get_config_option(
            "telescope_theta",
            self.config.zenith_angle + self.config.off_axis_angle,
        )
        command += super().get_config_option("star_photons", str(self.photons_per_run))
        command += super().get_config_option("telescope_phi", "0")
        command += super().get_config_option("camera_transmission", "1.0")
        command += super().get_config_option("nightsky_background", "all:0.")
        command += super().get_config_option("trigger_current_limit", "1e10")
        command += super().get_config_option("telescope_random_angle", "0")
        command += super().get_config_option("telescope_random_error", "0")
        command += super().get_config_option("convergent_depth", "0")
        command += super().get_config_option("maximum_telescopes", "1")
        command += super().get_config_option("show", "all")
        command += super().get_config_option("camera_filter", "none")
        if self.config.single_mirror_mode:
            command += super().get_config_option("focus_offset", "all:0.")
            command += super().get_config_option("camera_config_file", "single_pixel_camera.dat")
            command += super().get_config_option("camera_pixels", "1")
            command += super().get_config_option("trigger_pixels", "1")
            command += super().get_config_option("camera_body_diameter", "0")
            command += super().get_config_option(
                "mirror_list",
                self.telescope_model.get_single_mirror_list_file(
                    self.config.mirror_numbers, self.config.use_random_focal_length
                ),
            )
            command += super().get_config_option(
                "focal_length", self.config.source_distance * u.km.to(u.cm)
            )
            command += super().get_config_option("dish_shape_length", _mirror_focal_length)
            command += super().get_config_option("mirror_focal_length", _mirror_focal_length)
            command += super().get_config_option("parabolic_dish", "0")
            command += super().get_config_option("mirror_align_random_distance", "0.")
            command += super().get_config_option("mirror_align_random_vertical", "0.,28.,0.,0.")
        command += " " + str(self._corsika_file)

        return clear_default_sim_telarray_cfg_directories(command), self._log_file, self._log_file

    def _check_run_result(self, run_number=None):  # pylint: disable=unused-argument
        """
        Check run results.

        Photon list files should have at least 100 lines.

        Returns
        -------
        bool
            True if photon list is not empty.

        Raises
        ------
        RuntimeError
            if Photon list is empty.
        """
        with open(self._photons_file, "rb") as ff:
            n_lines = sum(1 for _ in ff)
        if n_lines < 100:
            raise RuntimeError("Photon list is empty.")
        return True

    def _write_out_single_pixel_camera_file(self):
        """Write out the single pixel camera file."""
        with self.telescope_model.config_file_directory.joinpath("single_pixel_camera.dat").open(
            "w"
        ) as file:
            file.write("# Single pixel camera\n")
            file.write('PixType 1   0  0 300   1 300 0.00   "funnel_perfect.dat"\n')
            file.write("Pixel 0 1 0. 0.  0  0  0 0x00 1\n")
            file.write("Trigger 1 of 0\n")

        # need to also write out the funnel_perfect.dat file
        with self.telescope_model.config_file_directory.joinpath("funnel_perfect.dat").open(
            "w"
        ) as file:
            file.write(
                "# Perfect light collection where the angular efficiency of funnels is needed\n"
            )
            file.write("0    1.0\n")
            file.write("30   1.0\n")
            file.write("60   1.0\n")
            file.write("90   1.0\n")

    def _config_to_namedtuple(self, data_dict):
        """Convert dict to namedtuple for configuration."""
        config_data = namedtuple(
            "Config",
            [
                "zenith_angle",
                "off_axis_angle",
                "source_distance",
                "single_mirror_mode",
                "use_random_focal_length",
                "mirror_numbers",
            ],
        )
        return config_data(
            zenith_angle=data_dict["zenith_angle"],
            off_axis_angle=data_dict["off_axis_angle"],
            source_distance=data_dict["source_distance"],
            single_mirror_mode=data_dict["single_mirror_mode"],
            use_random_focal_length=data_dict["use_random_focal_length"],
            mirror_numbers=data_dict["mirror_numbers"],
        )



#!/usr/bin/python3
r"""
    Convert all simulation model parameters exported from sim_telarray format.

    Check value, type, and range, convert units using schema files. Write json files
    ready to be submitted to the model database. Prints out parameters which are not found
    in simtel configuration file and parameters which are not found in simtools schema files.

    Note that all parameters are assigned the same parameter version.

    Command line arguments
    ----------------------
    simtel_cfg_file (str)
        File name of sim_telarray configuration file containing all simulation model parameters.

    simtel_telescope_name (str)
        Name of the telescope in the sim_telarray configuration file.

    telescope (str, optional)
        Telescope model name (e.g. LST-1, SST-D, ...)

    skip_parameter (str, optional)
        List of parameters to be skipped (use sim_telarray names).

    Example
    -------

    To export the model parameters from sim_telarray, first copy and unpack the configuration
    tar ball from sim_telarray (usually called 'sim_telarray_config.tar.gz') to the sim_telarray
    working directory. Extract the configuration using the following command:

    .. code-block:: console

        ./sim_telarray/bin/sim_telarray -c sim_telarray/cfg/CTA/CTA-PROD6-LaPalma.cfg \\
            -C limits=no-internal -C initlist=no-internal -C list=no-internal \\
            -C typelist=no-internal -C maximum_telescopes=30 -DNSB_AUTOSCALE \\
            -DNECTARCAM -DHYPER_LAYOUT -DNUM_TELESCOPES=30 /dev/null \\
            2>|/dev/null | grep '(@cfg)' | sed 's/^(@cfg) //' >| all_telescope_config_la_palma.cfg

        ./sim_telarray/bin/sim_telarray -c sim_telarray/cfg/CTA/CTA-PROD6-Paranal.cfg \\
            -C limits=no-internal -C initlist=no-internal -C list=no-internal \\
            -C typelist=no-internal -C maximum_telescopes=87 -DNSB_AUTOSCALE \\
            -DFLASHCAM -DHYPER_LAYOUT -DNUM_TELESCOPES=87 /dev/null \\
            2>|/dev/null | grep '(@cfg)' | sed 's/^(@cfg) //' >| all_telescope_config_paranal.cfg


    Extract then model parameters from the simtel_array configuration file for LSTN-01
    (telescopes are named CT1, CT2, ..., in the sim_telarray configuration file and must be
    provided in the "simtel_telescope_name" command line argument)
    and write json files in the same format as the model parameter database:

    .. code-block:: console

       simtools-convert-all-model-parameters-from-simtel \\
          --simtel_cfg_file all_telescope_config_la_palma.cfg\\
          --simtel_telescope_name CT1\\
          --telescope LSTN-01\\
          --parameter_version "1.0.0"\\
          --output_path /path/to/output

"""

import logging
from pathlib import Path

import numpy as np

import simtools.data_model.model_data_writer as writer
import simtools.utils.general as gen
from simtools.configuration import configurator
from simtools.data_model import schema
from simtools.io_operations.io_handler import IOHandler
from simtools.simtel import simtel_config_reader


def _parse(label=None, description=None):
    """
    Parse command line configuration.

    Parameters
    ----------
    label: str
        Label describing application.
    description: str
        Description of application.

    Returns
    -------
    CommandLineParser
        Command line parser object

    """
    config = configurator.Configurator(label=label, description=description)

    config.parser.add_argument(
        "--simtel_cfg_file",
        help="File name for simtel_array configuration",
        type=str,
        required=True,
    )
    config.parser.add_argument(
        "--simtel_telescope_name",
        help="Name of the telescope in the sim_telarray configuration file",
        type=str,
        required=True,
    )
    config.parser.add_argument(
        "--skip_parameter",
        help="List of parameters to be skipped.",
        type=str,
        nargs="*",
        default=[],
    )
    return config.initialize(simulation_model=["telescope", "parameter_version"])


def read_simtel_config_file(args_dict, schema_file, camera_pixels=None):
    """
    Read the simtel configuration file.

    Parameters
    ----------
    args_dict: dict
        Dictionary with command line arguments.
    schema_file: str
        Schema path name.
    camera_pixels: int
        Number of camera pixels.

    Returns
    -------
    SimtelConfigReader, None
        SimtelConfigReader object (None if parameter not found)

    """
    config_reader = simtel_config_reader.SimtelConfigReader(
        schema_file=schema_file,
        simtel_config_file=args_dict["simtel_cfg_file"],
        simtel_telescope_name=args_dict["simtel_telescope_name"],
        camera_pixels=camera_pixels,
    )
    if config_reader.parameter_dict is None or len(config_reader.parameter_dict) == 0:
        return None
    return config_reader


def get_number_of_camera_pixel(args_dict, logger):
    """
    Get the number of camera pixels from the simtel configuration file.

    Required to set the dimension some of the parameter correctly, as simtel
    in some cases does not provide the dimension ('all:' in the parameter files).

    Parameters
    ----------
    args_dict: dict
        Dictionary with command line arguments.
    logger: logging.Logger
        Logger object

    Returns
    -------
    int, None
        Number of camera pixels (None if file is not found)

    """
    try:
        config_reader = read_simtel_config_file(
            args_dict, schema.get_model_parameter_schema_file("camera_pixels")
        )
        _camera_pixel = config_reader.parameter_dict.get(args_dict["simtel_telescope_name"])
    except (FileNotFoundError, AttributeError):
        logger.warning("Failed to read camera pixel parameter.")
        _camera_pixel = None
    logger.info(f"Number of camera pixels: {_camera_pixel}")
    return _camera_pixel


def read_and_export_parameters(args_dict, logger):
    """
    Read and export parameters from simtel configuration file to json files.

    Provide extensive logging information on the parameters found in the simtel
    configuration file.

    Parameters
    ----------
    args_dict: dict
        Dictionary with command line arguments.
    logger: logging.Logger
        Logger object

    Returns
    -------
    list
        List of simtel parameters not found in schema files.
    list
        List of simtools parameter not found in simtel configuration file.

    """
    _parameters, _schema_files = schema.get_get_model_parameter_schema_files()
    _simtel_parameters = simtel_config_reader.get_list_of_simtel_parameters(
        args_dict["simtel_cfg_file"]
    )
    logger.info(f"Found {len(_simtel_parameters)} parameters in simtel configuration file.")

    io_handler = IOHandler()
    io_handler.set_paths(output_path=args_dict["output_path"])

    _camera_pixel = get_number_of_camera_pixel(args_dict, logger)

    _parameters_not_in_simtel = []

    for _parameter, _schema_file in zip(_parameters, _schema_files):
        logger.info(f"Parameter: {_parameter} Schema file: {_schema_file}")
        if _parameter in args_dict["skip_parameter"]:
            logger.info(f"Skipping {_parameter}")
            continue
        config_reader = read_simtel_config_file(args_dict, _schema_file, _camera_pixel)

        if config_reader is None:
            logger.info("Parameter not found in sim_telarray configuration file.")
            _parameters_not_in_simtel.append(_parameter)
            continue

        logger.info(f"Simtel parameter: {config_reader.parameter_dict}")

        _json_dict = writer.ModelDataWriter.dump_model_parameter(
            parameter_name=_parameter,
            value=config_reader.parameter_dict.get(args_dict["simtel_telescope_name"]),
            instrument=args_dict["telescope"],
            parameter_version=args_dict["parameter_version"],
            output_file=io_handler.get_output_file(
                f"{_parameter}-{args_dict['parameter_version']}.json",
                label=f"{args_dict['telescope']}",
                sub_dir=f"{_parameter}",
            ),
        )

        config_reader.compare_simtel_config_with_schema()

        if config_reader.simtel_parameter_name.lower() in _simtel_parameters:
            _simtel_parameters.remove(config_reader.simtel_parameter_name.lower())

        if _json_dict["file"]:
            logger.info(f"File name for {_parameter} is {_json_dict['value']}")

    return _parameters_not_in_simtel, _simtel_parameters


def print_parameters_not_found(_parameters_not_in_simtel, _simtel_parameters, args_dict, logger):
    """
    Print simtel/simtools parameter not found in schema and configuration files.

    For simtel parameters not found, check if the setting for the chose
    telescope is different from the default values.

    Parameters
    ----------
    _parameters_not_in_simtel: list
        List of simtel parameters not found in schema files.
    _simtel_parameters: list
        List of simtel parameters not found in simtools schema files.
    args_dict: dict
        Dictionary with command line arguments.
    logger: logging.Logger
        Logger object

    """
    logger.info(
        f"Parameters not found in simtools schema files ({len(_parameters_not_in_simtel)}):"
    )
    for para in sorted(_parameters_not_in_simtel):
        logger.info(f"  {para}")

    logger.info(f"Simtel parameters not found in schema files ({len(_simtel_parameters)}):")
    for para in sorted(_simtel_parameters):
        logger.info(f"Simtel parameter: {para}")
        config_reader = simtel_config_reader.SimtelConfigReader(
            schema_file=None,
            simtel_config_file=args_dict["simtel_cfg_file"],
            simtel_telescope_name=args_dict["simtel_telescope_name"],
            parameter_name=para,
        )
        _default = config_reader.parameter_dict.get("default")
        _tel_value = config_reader.parameter_dict.get(args_dict["simtel_telescope_name"])
        # simple comparison of default value and telescope values, does not work for lists
        try:
            if _default == _tel_value or np.isclose(_default, _tel_value):
                logger.info(f"    Default and telescope values for {para} are equal: {_default}")
                continue
        except (ValueError, TypeError):
            pass
        if isinstance(_default, np.ndarray):
            logger.warning(f"    Default value ({para}): {_default} (length: {_default.size})")
        else:
            logger.warning(f"    Default value ({para}): {_default}")
        if isinstance(_tel_value, np.ndarray):
            logger.warning(
                f"    Telescope value ({para}): {_tel_value} (length: {_tel_value.size})"
            )
        else:
            logger.warning(f"    Telescope value ({para}): {_tel_value}")


def print_list_of_files(args_dict, logger):
    """
    Print model parameters which describe a file name.

    This is useful to find files which are part of the model.

    Parameters
    ----------
    args_dict: dict
        Dictionary with command line arguments.
    logger: logging.Logger
        Logger object

    """
    model_files = sorted(Path(args_dict["output_path"]).rglob("*.json"))
    for file in model_files:
        model_dict = gen.collect_data_from_file(file_name=file)
        if model_dict.get("file"):
            logger.info(f"{file.name}: {model_dict['value']}")


def main():  # noqa: D103
    args_dict, _ = _parse(
        label=Path(__file__).stem,
        description="Convert simulation model parameters from sim_telarray to simtools format.",
    )
    logger = logging.getLogger()
    logger.setLevel(gen.get_log_level_from_user(args_dict["log_level"]))

    _parameters_not_in_simtel, _simtel_parameters = read_and_export_parameters(args_dict, logger)
    print_parameters_not_found(_parameters_not_in_simtel, _simtel_parameters, args_dict, logger)
    print_list_of_files(args_dict, logger)


if __name__ == "__main__":
    main()



"""Run simulations with CORSIKA and pipe it to sim_telarray using the multipipe functionality."""

import logging
import stat
from pathlib import Path

from simtools.runners.corsika_runner import CorsikaRunner
from simtools.simtel.simulator_array import SimulatorArray
from simtools.utils.general import clear_default_sim_telarray_cfg_directories

__all__ = ["CorsikaSimtelRunner"]


class CorsikaSimtelRunner:
    """
    Run simulations with CORSIKA and pipe it to sim_telarray using the multipipe functionality.

    Uses CorsikaConfig to manage the CORSIKA configuration and SimulatorArray
    for the sim_telarray configuration.

    Parameters
    ----------
    corsika_config : CorsikaConfig
        Configuration parameters for CORSIKA.
    simtel_path : str or Path
        Location of the sim_telarray package.
    label : str
        Label.
    keep_seeds : bool
        Use seeds based on run number and primary particle. If False, use sim_telarray seeds.
    use_multipipe : bool
        Use multipipe to run CORSIKA and sim_telarray.
    """

    def __init__(
        self,
        corsika_config,
        simtel_path,
        label=None,
        keep_seeds=False,
        use_multipipe=False,
        sim_telarray_seeds=None,
    ):
        self._logger = logging.getLogger(__name__)
        self.corsika_config = corsika_config
        self._simtel_path = simtel_path
        self.sim_telarray_seeds = sim_telarray_seeds
        self.label = label

        self.corsika_config.set_output_file_and_directory(use_multipipe)
        self.corsika_runner = CorsikaRunner(
            corsika_config=corsika_config,
            simtel_path=simtel_path,
            label=label,
            keep_seeds=keep_seeds,
            use_multipipe=use_multipipe,
        )
        self.simulator_array = SimulatorArray(
            corsika_config=corsika_config,
            simtel_path=simtel_path,
            label=label,
            use_multipipe=use_multipipe,
            sim_telarray_seeds=sim_telarray_seeds,
        )

    def prepare_run_script(
        self, run_number=None, input_file=None, extra_commands=None, use_pfp=False
    ):
        """
        Get the full path of the run script file for a given run number.

        Parameters
        ----------
        run_number: int
            Run number.
        use_pfp: bool
            Whether to use the preprocessor in preparing the CORSIKA input file

        Returns
        -------
        Path:
            Full path of the run script file.
        """
        self._export_multipipe_script(run_number)
        return self.corsika_runner.prepare_run_script(
            run_number=run_number,
            input_file=input_file,
            extra_commands=extra_commands,
            use_pfp=use_pfp,
        )

    def _export_multipipe_script(self, run_number):
        """
        Write the multipipe script used in piping CORSIKA to sim_telarray.

        Parameters
        ----------
        run_number: int
            Run number.

        Returns
        -------
        Path:
            Full path of the run script file.
        """
        run_command = self._make_run_command(
            run_number=run_number,
            input_file="-",  # instruct sim_telarray to take input from standard output
        )
        multipipe_file = Path(self.corsika_config.config_file_path.parent).joinpath(
            self.corsika_config.get_corsika_config_file_name("multipipe")
        )
        with open(multipipe_file, "w", encoding="utf-8") as file:
            file.write(f"{run_command}")
        self._logger.info(f"Multipipe script: {multipipe_file}")
        self._write_multipipe_script(multipipe_file)

    def _write_multipipe_script(self, multipipe_file):
        """
        Write script used to call the multipipe_corsika command.

        Parameters
        ----------
        multipipe_file: str or Path
            The name of the multipipe file which contains all of the multipipe commands.
        """
        multipipe_script = Path(self.corsika_config.config_file_path.parent).joinpath(
            "run_cta_multipipe"
        )
        with open(multipipe_script, "w", encoding="utf-8") as file:
            multipipe_command = Path(self._simtel_path).joinpath(
                f"sim_telarray/bin/multipipe_corsika -c {multipipe_file} || echo 'Fan-out failed'"
            )
            file.write(f"{multipipe_command}")

        multipipe_script.chmod(multipipe_script.stat().st_mode | stat.S_IEXEC)

    def _make_run_command(self, run_number=None, input_file=None):
        """
        Build and return the command to run simtel_array.

        Parameters
        ----------
        run_number: int
            Run number.
        input_file: str
            Full path of the input CORSIKA file.
            Use '-' to tell sim_telarray to read from standard output

        Returns
        -------
        str:
            Command to run sim_telarray.
        """
        try:
            weak_pointing = any(pointing in self.label for pointing in ["divergent", "convergent"])
        except TypeError:  # allow for self.label to be None
            weak_pointing = False

        command = str(self._simtel_path.joinpath("sim_telarray/bin/sim_telarray"))
        command += f" -c {self.corsika_config.array_model.get_config_file()}"
        command += f" -I{self.corsika_config.array_model.get_config_directory()}"
        command += self.simulator_array.get_config_option(
            "telescope_theta", self.corsika_config.zenith_angle, weak_option=weak_pointing
        )
        command += self.simulator_array.get_config_option(
            "telescope_phi", self.corsika_config.azimuth_angle, weak_option=weak_pointing
        )
        command += self.simulator_array.get_config_option(
            "power_law",
            SimulatorArray.get_power_law_for_sim_telarray_histograms(
                self.corsika_config.primary_particle
            ),
        )
        command += self.simulator_array.get_config_option(
            "histogram_file",
            self.get_file_name(
                simulation_software="simtel", file_type="histogram", run_number=run_number
            ),
        )
        command += self.simulator_array.get_config_option("random_state", "none")
        if self.sim_telarray_seeds:
            command += self.simulator_array.get_config_option(
                "random_seed", self.sim_telarray_seeds
            )
        command += self.simulator_array.get_config_option("show", "all")
        command += self.simulator_array.get_config_option(
            "output_file",
            self.simulator_array.get_file_name(
                simulation_software="simtel", file_type="output", run_number=run_number
            ),
        )
        command += f" {input_file}"
        _log_file = self.simulator_array.get_file_name(
            simulation_software="simtel", file_type="log", run_number=run_number
        )
        command += f" | gzip > {_log_file} 2>&1 || exit"

        # Remove the default sim_telarray configuration directories
        return clear_default_sim_telarray_cfg_directories(command)

    def get_file_name(self, simulation_software=None, file_type=None, run_number=None, mode=None):
        """
        Get the full path of a file for a given run number.

        Parameters
        ----------
        simulation_software: str
            Simulation software.
        file_type: str
            File type.
        run_number: int
            Run number.

        Returns
        -------
        str
            File name with full path.
        """
        if simulation_software is None:
            # preference to simtel output (multipipe)
            simulation_software = "simtel" if self.simulator_array else "corsika"

        runner = self.corsika_runner if simulation_software == "corsika" else self.simulator_array
        return runner.get_file_name(file_type=file_type, run_number=run_number, mode=mode)



#!/usr/bin/python3
"""Plot tabular data."""

import numpy as np
from astropy.table import Table

import simtools.utils.general as gen
from simtools.db import db_handler
from simtools.io_operations import legacy_data_handler
from simtools.visualization import visualize


def plot(config, output_file, db_config=None):
    """
    Plot tabular data from data or from model parameter files.

    Parameters
    ----------
    config: dict
        Configuration dictionary for plotting.
    output_file: str
        Output file.
    """
    data = read_table_data(config, db_config)

    fig = visualize.plot_1d(
        data,
        **config,
    )
    visualize.save_figure(fig, output_file)


def read_table_data(config, db_config):
    """
    Read table data from file or parameter database.

    Parameters
    ----------
    config: dict
        Configuration dictionary for plotting.

    Returns
    -------
    Dict
        Dict with table data (astropy tables).
    """
    data = {}

    for _config in config["tables"]:
        if "parameter" in _config:
            table = _read_table_from_model_database(_config, db_config)
        elif "file_name" in _config:
            if "legacy" in _config.get("type", ""):
                table = legacy_data_handler.read_legacy_data_as_table(
                    _config["file_name"], _config["type"]
                )
            else:
                table = Table.read(_config["file_name"], format="ascii.ecsv")
        else:
            raise ValueError("No table data defined in configuration.")

        if _config.get("normalize_y"):
            table[_config["column_y"]] = (
                table[_config["column_y"]] / table[_config["column_y"]].max()
            )
        if _config.get("select_values"):
            table = _select_values_from_table(
                table,
                _config["select_values"]["column_name"],
                _config["select_values"]["value"],
            )
        data[_config["label"]] = gen.get_structure_array_from_table(
            table,
            [
                _config["column_x"],
                _config["column_y"],
                _config.get("column_x_err"),
                _config.get("column_y_err"),
            ],
        )
    return data


def _read_table_from_model_database(table_config, db_config):
    """
    Read table data from model parameter database.

    Parameters
    ----------
    table_config: dict
        Configuration dictionary for table data.

    Returns
    -------
    Table
        Astropy table.
    """
    db = db_handler.DatabaseHandler(mongo_db_config=db_config)
    return db.export_model_file(
        parameter=table_config["parameter"],
        site=table_config["site"],
        array_element_name=table_config.get("telescope"),
        parameter_version=table_config.get("parameter_version"),
        model_version=table_config.get("model_version"),
        export_file_as_table=True,
    )


def _select_values_from_table(table, column_name, value):
    """Return a table with only the rows where column_name == value."""
    return table[np.isclose(table[column_name], value)]



#!/usr/bin/python3
"""
    Read array element positions from file and add them to model repository.

    This is an application for experts and should not be used by the general user.
    Reading of input is fine-tuned to the array element files as provided by CTAO.

    Command line arguments

    input : str
        File containing a table of array element positions.
    repository_path : str
        Path of local copy of model parameter repository.
    parameter_version : str
        Parameter version.
    site : str
        Observatory site.
    coordinate_system : str
        Coordinate system of array element positions (ground or utm).

    Examples
    --------
    Add array element positions to repository:

    .. code-block:: console

        simtools-write-array-element-positions-to-repository \
            --input /path/to/positions.txt \
            --repository_path /path/to/repository \
            --parameter_version 0.1.0 \
            --coordinate_system ground \
            --site North

"""

import json
import logging
from pathlib import Path

import astropy.table

import simtools.utils.general as gen
from simtools.configuration import configurator
from simtools.data_model.model_data_writer import JsonNumpyEncoder, ModelDataWriter
from simtools.model.array_model import ArrayModel


def _parse(label=None, description=None):
    """
    Parse command line configuration.

    Parameters
    ----------
    label : str
        Label describing application.
    description : str
        Description of application.

    Returns
    -------
    CommandLineParser
        Command line parser object.
    """
    config = configurator.Configurator(label=label, description=description)
    config.parser.add_argument(
        "--input",
        help="File containing a table of array element positions.",
        required=False,
    )
    config.parser.add_argument(
        "--repository_path",
        help="Output path to model parameter repository.",
        type=Path,
        required=False,
    )
    config.parser.add_argument(
        "--coordinate_system",
        help="Coordinate system of array element positions (utm or ground).",
        default="ground",
        required=False,
        type=str,
        choices=["ground", "utm"],
    )

    return config.initialize(db_config=True, simulation_model=["site", "parameter_version"])


def write_utm_array_elements_to_repository(args_dict, logger):
    """
    Write UTM position of array elements to model repository.

    Read array element positions from file. The ecsv row definition might
    include telescope_name or asset_code and sequence_number.

    Parameters
    ----------
    args_dict : dict
        Command line arguments.
    logger : Logger
        Logger object.

    """
    array_elements = astropy.table.Table.read(args_dict["input"])
    for row in array_elements:
        instrument = (
            row["telescope_name"]
            if "telescope_name" in array_elements.colnames
            else f"{row['asset_code']}-{row['sequence_number']}"
        )
        output_path = Path(args_dict["repository_path"]) / f"{instrument}"
        output_path.mkdir(parents=True, exist_ok=True)
        logger.info(f"Writing array element positions (utm) to {output_path}")

        ModelDataWriter.dump_model_parameter(
            parameter_name="array_element_position_utm",
            instrument=instrument,
            value=f"{row['utm_east']} {row['utm_north']} {row['altitude']}",
            parameter_version=args_dict["parameter_version"],
            output_path=output_path,
            output_file="array_element_position_utm.json",
        )


def write_ground_array_elements_to_repository(args_dict, db_config, logger):
    """
    Write ground position of array elements to model repository.

    Parameters
    ----------
    args_dict : dict
        Command line arguments.
    db_config : dict
        Database configuration.
    logger : Logger
        Logger object.

    """
    array_model = ArrayModel(
        mongo_db_config=db_config,
        model_version=None,
        site=args_dict["site"],
        array_elements=args_dict["input"],
    )
    for element_name, data in array_model.array_elements.items():
        output_path = Path(args_dict["repository_path"]) / f"{element_name}"
        output_path.mkdir(parents=True, exist_ok=True)
        logger.info(f"Writing array element positions (ground) to {output_path}")
        with open(
            output_path / "array_element_position_ground.json", "w", encoding="utf-8"
        ) as file:
            json.dump(
                data,
                file,
                indent=4,
                sort_keys=False,
                cls=JsonNumpyEncoder,
            )
            file.write("\n")


def main():
    """Application main."""
    label = Path(__file__).stem
    args_dict, db_config = _parse(
        label, description="Add array element positions to model parameter repository"
    )
    logger = logging.getLogger()
    logger.setLevel(gen.get_log_level_from_user(args_dict["log_level"]))

    if args_dict["coordinate_system"] == "utm":
        write_utm_array_elements_to_repository(args_dict, logger)
    elif args_dict["coordinate_system"] == "ground":
        write_ground_array_elements_to_repository(args_dict, db_config, logger)
    else:
        logger.error("Invalid coordinate system. Allowed are 'utm' and 'ground'.")
        raise ValueError


if __name__ == "__main__":
    main()



#!/usr/bin/python3

r"""
    Add a new parameter / value to a collection in the DB using a json file as input.

    Command line arguments
    ----------------------
    file_name (str, required)
        Name of the file to upload including the full path.
    db_collection (str, required)
        The DB collection to which to add the file.
    db (str)
        The DB to insert the files to.

    Example
    -------

    Upload a file to sites collection:

    .. code-block:: console

        simtools-add-value-from-json-to-db \\
            --file_name new_value.json --db_collection sites


"""

import logging
from pathlib import Path

import simtools.utils.general as gen
from simtools.configuration import configurator
from simtools.db import db_handler


def _parse():
    config = configurator.Configurator(description="Add a new parameter to the DB.")
    group = config.parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--file_name", help="file to be added", type=str)
    group.add_argument(
        "--input_path",
        help="A directory with json files to upload to the DB.",
        type=Path,
    )
    config.parser.add_argument(
        "--db_collection", help="DB collection to which to add new values.", required=True
    )
    config.parser.add_argument("--db", help="Database name", type=str, required=True)
    return config.initialize(db_config=True)


def main():  # noqa: D103
    args_dict, db_config = _parse()

    logger = logging.getLogger()
    logger.setLevel(gen.get_log_level_from_user(args_dict["log_level"]))

    db = db_handler.DatabaseHandler(mongo_db_config=db_config)

    files_to_insert = []
    if args_dict.get("file_name", None) is not None:
        files_to_insert.append(args_dict["file_name"])
    else:
        files_to_insert.extend(Path(args_dict["input_path"]).glob("*json"))

    if len(files_to_insert) < 1:
        raise ValueError("No files were provided to upload")
    plural = "s" if len(files_to_insert) > 1 else ""

    print(
        f"Should the following parameter{plural} be inserted to the "
        f"{args_dict['db_collection']} DB collection?:\n"
    )
    print(*files_to_insert, sep="\n")
    print()

    if gen.user_confirm():
        for file_to_insert_now in files_to_insert:
            par_dict = gen.collect_data_from_file(file_name=file_to_insert_now)
            logger.info(f"Adding the following parameter to the DB: {par_dict['parameter']}")
            db.add_new_parameter(
                db_name=db_config["db_simulation_model"],
                par_dict=par_dict,
                collection_name=args_dict["db_collection"],
                file_prefix="./",
            )
            logger.info(
                f"Value for {par_dict['parameter']} added to "
                f"{args_dict['db_collection']} collection."
            )
    else:
        logger.info("Aborted, no change applied to the database")


if __name__ == "__main__":
    main()



"""Camera efficiency simulations and analysis."""

import logging
import re
from collections import defaultdict

import astropy.io.ascii
import numpy as np
from astropy.table import Table

from simtools.io_operations import io_handler
from simtools.model.site_model import SiteModel
from simtools.model.telescope_model import TelescopeModel
from simtools.simtel.simulator_camera_efficiency import SimulatorCameraEfficiency
from simtools.utils import names
from simtools.visualization import visualize

__all__ = ["CameraEfficiency"]


class CameraEfficiency:
    """
    Camera efficiency simulations and analysis.

    Parameters
    ----------
    simtel_path: str (or Path)
        Location of sim_telarray installation.
    db_config: dict
        Configuration for the database.
    label: str
        Instance label, optional.
    config_data: dict.
        Dict containing the configurable parameters.
    test: bool
        Is it a test instance (at the moment only affects the location of files).
    """

    def __init__(
        self,
        simtel_path,
        config_data,
        label,
        db_config,
        test=False,
    ):
        """Initialize the CameraEfficiency class."""
        self._logger = logging.getLogger(__name__)

        self._simtel_path = simtel_path
        self.label = label
        self.test = test

        self.io_handler = io_handler.IOHandler()
        self.telescope_model, self._site_model = self._initialize_simulation_models(
            config_data, db_config
        )
        self.output_dir = self.io_handler.get_output_directory(self.label, sub_dir="plots")

        self._results = None
        self._has_results = False

        self.config = self._configuration_from_args_dict(config_data)
        self._file = self._load_files()

    def __repr__(self):
        """Return string representation of the CameraEfficiency instance."""
        return f"CameraEfficiency(label={self.label})\n"

    def _initialize_simulation_models(self, config_data, db_config):
        """
        Initialize site and telescope models.

        Parameters
        ----------
        config_data: dict
            Dict containing the configurable parameters.

        Returns
        -------
        tuple
            Tuple containing the site and telescope models.
        """
        tel_model = TelescopeModel(
            site=config_data["site"],
            telescope_name=config_data["telescope"],
            mongo_db_config=db_config,
            model_version=config_data["model_version"],
            label=self.label,
        )
        site_model = SiteModel(
            site=config_data["site"],
            model_version=config_data["model_version"],
            mongo_db_config=db_config,
        )
        return tel_model, site_model

    def _configuration_from_args_dict(self, config_data):
        """
        Extract the configuration data from the args_dict.

        Zenith and azimuth angles are set to default values if not provided.

        Parameters
        ----------
        config_data: dict
            Dict containing the configurable parameters.

        Returns
        -------
        dict
            Configuration data.
        """
        zenith_angle = config_data.get("zenith_angle")
        if zenith_angle is not None:
            zenith_angle = zenith_angle.to("deg").value
        else:
            zenith_angle = 20.0
            self._logger.info(f"Setting zenith angle to default value {zenith_angle} deg")
        azimuth_angle = config_data.get("azimuth_angle")
        if azimuth_angle is not None:
            azimuth_angle = azimuth_angle.to("deg").value
        else:
            azimuth_angle = 0.0
            self._logger.info(f"Setting azimuth angle to default value {azimuth_angle} deg")

        return {
            "zenith_angle": zenith_angle,
            "azimuth_angle": azimuth_angle,
            "nsb_spectrum": config_data.get("nsb_spectrum", None),
        }

    def _load_files(self):
        """Define the variables for the file names, including the results, simtel and log file."""
        _file = {}
        for label, suffix in zip(
            ["results", "simtel", "log"],
            [".ecsv", ".dat", ".log"],
        ):
            file_name = names.generate_file_name(
                file_type=(
                    "camera-efficiency-table" if label == "results" else "camera-efficiency"
                ),
                suffix=suffix,
                site=self.telescope_model.site,
                telescope_model_name=self.telescope_model.name,
                zenith_angle=self.config["zenith_angle"],
                azimuth_angle=self.config["azimuth_angle"],
                label=self.label,
            )

            _file[label] = self.io_handler.get_output_directory(
                label=self.label,
                sub_dir="camera-efficiency",
            ).joinpath(file_name)
        return _file

    def simulate(self):
        """Simulate camera efficiency using testeff."""
        self._logger.info("Simulating CameraEfficiency")

        self.export_model_files()

        simtel = SimulatorCameraEfficiency(
            simtel_path=self._simtel_path,
            telescope_model=self.telescope_model,
            zenith_angle=self.config["zenith_angle"],
            file_simtel=self._file["simtel"],
            file_log=self._file["log"],
            label=self.label,
            nsb_spectrum=self.config["nsb_spectrum"],
            skip_correction_to_nsb_spectrum=self.config.get(
                "skip_correction_to_nsb_spectrum", False
            ),
        )
        simtel.run(test=self.test)

    def export_model_files(self):
        """Export model and config files to the output directory."""
        self.telescope_model.export_config_file()
        self.telescope_model.export_model_files()
        if not self.config.get("skip_correction_to_nsb_spectrum", False):
            self.telescope_model.export_nsb_spectrum_to_telescope_altitude_correction_file(
                model_directory=self.telescope_model.config_file_directory
            )

    def analyze(self, export=True, force=False):
        """
        Analyze camera efficiency output file and store the results in _results.

        Parameters
        ----------
        export: bool
            If True, results will be exported to a file automatically. Alternatively, export_results
            function can be used.
        force: bool
            If True, existing results files will be removed and analysis will be done again.
        """
        self._logger.info("Analyzing CameraEfficiency")

        if "results" in self._file and not force:
            self._logger.info("Results file exists and force=False - skipping analyze")
            self._read_results()
            return

        # List of parameters to be calculated and stored
        eff_pars = [
            "wl",
            "eff",
            "eff_atm",
            "qe",
            "ref",
            "masts",
            "filt",
            "pixel",
            "atm_trans",
            "cher",
            "nsb",
            "atm_corr",
            "nsb_site",
            "nsb_site_eff",
            "nsb_be",
            "nsb_be_eff",
            "C1",
            "C2",
            "C3",
            "C4",
            "C4x",
            "N1",
            "N2",
            "N3",
            "N4",
            "N4x",
        ]

        _results = defaultdict(list)

        # Search for at least 5 consecutive numbers to see that we are in the table
        re_table = re.compile("{0}{0}{0}{0}{0}".format(r"[-+]?[0-9]*\.?[0-9]+\s+"))
        with open(self._file["simtel"], encoding="utf-8") as file:
            for line in file:
                if re_table.match(line):
                    words = line.split()
                    numbers = [float(w) for w in words]
                    for i in range(len(eff_pars) - 10):
                        _results[eff_pars[i]].append(numbers[i])
                    C1 = numbers[8] * (400 / numbers[0]) ** 2  # noqa: N806
                    C2 = C1 * numbers[4] * numbers[5]  # noqa: N806
                    C3 = C2 * numbers[6] * numbers[7]  # noqa: N806
                    C4 = C3 * numbers[3]  # noqa: N806
                    c4x_value = C1 * numbers[3] * numbers[6] * numbers[7]
                    _results["C1"].append(C1)
                    _results["C2"].append(C2)
                    _results["C3"].append(C3)
                    _results["C4"].append(C4)
                    _results["C4x"].append(c4x_value)
                    N1 = numbers[14]  # noqa: N806
                    N2 = N1 * numbers[4] * numbers[5]  # noqa: N806
                    N3 = N2 * numbers[6] * numbers[7]  # noqa: N806
                    N4 = N3 * numbers[3]  # noqa: N806
                    n4x_value = N1 * numbers[3] * numbers[6] * numbers[7]
                    _results["N1"].append(N1)
                    _results["N2"].append(N2)
                    _results["N3"].append(N3)
                    _results["N4"].append(N4)
                    _results["N4x"].append(n4x_value)

        self._results = Table(_results)
        self._has_results = True

        print("\33[40;37;1m")
        self._logger.info(f"\n{self.results_summary()}")
        print("\033[0m")

        if export:
            self.export_results()

    def results_summary(self):
        """
        Print a summary of the results.

        Include a header for the zenith/azimuth settings and the NSB spectrum file which was used.
        The summary includes the various CTAO requirements and the final expected NSB pixel rate.
        """
        nsb_pixel_pe_per_ns, nsb_rate_ref_conditions = self.calc_nsb_rate()
        nsb_spectrum_text = (
            f"NSB spectrum file: {self.config['nsb_spectrum']}"
            if self.config["nsb_spectrum"]
            else "default sim_telarray spectrum."
        )
        return (
            f"Results summary for {self.telescope_model.name} at "
            f"zenith={self.config['zenith_angle']:.1f} deg, "
            f"azimuth={self.config['azimuth_angle']:.1f} deg\n"
            f"Using the {nsb_spectrum_text}\n"
            f"\nSpectrum weighted reflectivity: {self.calc_reflectivity():.4f}\n"
            "Camera nominal efficiency with gaps (B-TEL-1170): "
            f"{self.calc_camera_efficiency():.4f}\n"
            "Telescope total efficiency"
            f" with gaps (was A-PERF-2020): {self.calc_tel_efficiency():.4f}\n"
            "Telescope total Cherenkov light efficiency / sqrt(total NSB efficiency) "
            "(A-PERF-2025/B-TEL-0090): "
            f"{self.calc_tot_efficiency(self.calc_tel_efficiency()):.4f}\n"
            "Expected NSB pixel rate for the provided NSB spectrum: "
            f"{nsb_pixel_pe_per_ns:.4f} [p.e./ns]\n"
            "Expected NSB pixel rate for the reference NSB: "
            f"{nsb_rate_ref_conditions:.4f} [p.e./ns]\n"
        )

    def export_results(self):
        """Export results to a ecsv file."""
        if not self._has_results:
            self._logger.error("Cannot export results because they do not exist")
        else:
            self._logger.info(f"Exporting testeff table to {self._file['results']}")
            astropy.io.ascii.write(
                self._results, self._file["results"], format="basic", overwrite=True
            )
            _results_summary_file = (
                str(self._file["results"]).replace(".ecsv", ".txt").replace("-table-", "-summary-")
            )
            self._logger.info(f"Exporting summary results to {_results_summary_file}")
            with open(_results_summary_file, "w", encoding="utf-8") as file:
                file.write(self.results_summary())

    def _read_results(self):
        """Read existing results file and store it in _results."""
        table = astropy.io.ascii.read(self._file["results"], format="basic")
        self._results = table
        self._has_results = True

    def calc_tel_efficiency(self):
        """
        Calculate the telescope total efficiency including gaps (as defined in A-PERF-2020).

        Returns
        -------
        tel_efficiency: float
            Telescope efficiency
        """
        # Sum(C1) from 300 - 550 nm:
        c1_reduced_wl = self._results["C1"][[299 < wl_now < 551 for wl_now in self._results["wl"]]]
        c1_sum = np.sum(c1_reduced_wl)
        # Sum(C4) from 200 - 999 nm:
        c4_sum = np.sum(self._results["C4"])
        masts_factor = self._results["masts"][0]
        fill_factor = self.telescope_model.camera.get_camera_fill_factor()

        return fill_factor * (c4_sum / (masts_factor * c1_sum))

    def calc_camera_efficiency(self):
        """
        Calculate the camera nominal efficiency including gaps (as defined in B-TEL-1170).

        Returns
        -------
        cam_efficiency: float
            Wavelength-averaged camera efficiency
        """
        # Sum(C1) from 300 - 550 nm:
        c1_reduced_wl = self._results["C1"][[299 < wl_now < 551 for wl_now in self._results["wl"]]]
        c1_sum = np.sum(c1_reduced_wl)
        # Sum(C4x) from 300 - 550 nm:
        c4x_reduced_wl = self._results["C4x"][
            [299 < wl_now < 551 for wl_now in self._results["wl"]]
        ]
        c4x_sum = np.sum(c4x_reduced_wl)
        fill_factor = self.telescope_model.camera.get_camera_fill_factor()

        cam_efficiency_no_gaps = c4x_sum / c1_sum
        return cam_efficiency_no_gaps * fill_factor

    def calc_tot_efficiency(self, tel_efficiency):
        """
        Calculate the telescope total efficiency including gaps (as defined in A-PERF-2020).

        Parameters
        ----------
        tel_efficiency: float
            The telescope efficiency as calculated by calc_tel_efficiency()

        Returns
        -------
        Float
            Telescope total efficiency including gaps
        """
        # Sum(N1) from 300 - 550 nm:
        n1_reduced_wl = self._results["N1"][[299 < wl_now < 551 for wl_now in self._results["wl"]]]
        n1_sum = np.sum(n1_reduced_wl)
        # Sum(N4) from 200 - 999 nm:
        n4_sum = np.sum(self._results["N4"])
        masts_factor = self._results["masts"][0]
        fill_factor = self.telescope_model.camera.get_camera_fill_factor()

        tel_efficiency_nsb = fill_factor * (n4_sum / (masts_factor * n1_sum))

        return tel_efficiency / np.sqrt(tel_efficiency_nsb)

    def calc_reflectivity(self):
        """
        Calculate the Cherenkov spectrum weighted reflectivity in the range 300-550 nm.

        Returns
        -------
        Float
            Cherenkov spectrum weighted reflectivity (300-550 nm)
        """
        # Sum(C1) from 300 - 550 nm:
        c1_reduced_wl = self._results["C1"][[299 < wl_now < 551 for wl_now in self._results["wl"]]]
        c1_sum = np.sum(c1_reduced_wl)
        # Sum(C2) from 300 - 550 nm:
        c2_reduced_wl = self._results["C2"][[299 < wl_now < 551 for wl_now in self._results["wl"]]]
        c2_sum = np.sum(c2_reduced_wl)
        return c2_sum / c1_sum / self._results["masts"][0]

    def calc_nsb_rate(self):
        """
        Calculate the NSB rate.

        Returns
        -------
        nsb_rate_provided_spectrum: float
            NSB pixel rate in p.e./ns for the provided NSB spectrum
        nsb_rate_ref_conditions: float
            NSB pixel rate in p.e./ns for reference conditions
            (https://jama.cta-observatory.org/perspective.req#/items/26694?projectId=11)
        """
        nsb_rate_provided_spectrum = (
            np.sum(self._results["N4"])
            * self.telescope_model.camera.get_pixel_active_solid_angle()
            * self.telescope_model.get_on_axis_eff_optical_area().to("m2").value
            / self.telescope_model.get_parameter_value("telescope_transmission")[0]
        )

        # (integral is in ph./(m^2 ns sr) ) from 300 - 650 nm:
        n1_reduced_wl = self._results["N1"][[299 < wl_now < 651 for wl_now in self._results["wl"]]]
        n1_sum = np.sum(n1_reduced_wl)
        n1_integral_edges = self._results["N1"][
            [wl_now in [300, 650] for wl_now in self._results["wl"]]
        ]
        n1_integral_edges_sum = np.sum(n1_integral_edges)
        nsb_integral = 0.0001 * (n1_sum - 0.5 * n1_integral_edges_sum)
        nsb_rate_ref_conditions = (
            nsb_rate_provided_spectrum
            * self._site_model.get_parameter_value("nsb_reference_value")
            / nsb_integral
        )
        return nsb_rate_provided_spectrum, nsb_rate_ref_conditions

    def plot_efficiency(self, efficiency_type, save_fig=False):
        """
        Plot efficiency vs wavelength.

        Parameters
        ----------
        efficiency_type: str
            The type of efficiency to plot (Cherenkov 'C' or NSB 'N')
        save_fig: bool
            If True, the figure will be saved to a file.

        Returns
        -------
        fig
            The figure instance of pyplot
        """
        self._logger.info(f"Plotting {efficiency_type} efficiency vs wavelength")

        _col_type = "C" if efficiency_type == "Cherenkov" else "N"

        column_titles = {
            "wl": "Wavelength [nm]",
            f"{_col_type}1": rf"{_col_type}1: Cherenkov light on ground",
            f"{_col_type}2": rf"{_col_type}2: {_col_type}1 $\times$ ref. $\times$ masts",
            f"{_col_type}3": rf"{_col_type}3: {_col_type}2 $\times$ filter $\times$ lightguide",
            f"{_col_type}4": rf"{_col_type}4: {_col_type}3 $\times$ q.e.",
            f"{_col_type}4x": (
                rf"{_col_type}4x: {_col_type}1 $\times$ filter $\times$ lightguide $\times$ q.e."
            ),
        }

        table_to_plot = Table([self._results[col_now] for col_now in column_titles])

        for column_now, column_title in column_titles.items():
            table_to_plot.rename_column(column_now, column_title)

        y_title = f"{efficiency_type} light efficiency"
        if efficiency_type == "NSB":
            y_title = r"Diff. ph. rate [$10^{9} \times $ph/(nm s m$^2$ sr)]"
        plot = visualize.plot_table(
            table_to_plot,
            y_title=y_title,
            title=f"{self.telescope_model.name} response to {efficiency_type} light",
            no_markers=True,
        )
        if efficiency_type == "NSB":
            plot.gca().set_yscale("log")
            ylim = plot.gca().get_ylim()
            plot.gca().set_ylim(1e-3, ylim[1])
        if save_fig:
            self._save_plot(plot, efficiency_type.lower())
        return plot

    def _save_plot(self, fig, plot_title):
        """
        Save plot to pdf and png file.

        Parameters
        ----------
        fig
            The figure instance of pyplot
        plot_title: str
            The title of the plot
        """
        plot_file = self.output_dir.joinpath(
            self.label + "_" + self.telescope_model.name + "_" + plot_title
        )
        visualize.save_figure(fig, plot_file, log_title=f"{plot_title} efficiency")



"""Definition of the ArrayModel class."""

import logging
from pathlib import Path

import astropy.units as u
from astropy.table import QTable

from simtools.data_model import data_reader, schema
from simtools.db import db_handler
from simtools.io_operations import io_handler
from simtools.model.site_model import SiteModel
from simtools.model.telescope_model import TelescopeModel
from simtools.simtel.simtel_config_writer import SimtelConfigWriter
from simtools.utils import general, names

__all__ = ["ArrayModel"]


class ArrayModel:
    """
    Representation of an observatory consisting of site, telescopes, and further devices.

    Parameters
    ----------
    mongo_db_config: dict
        MongoDB configuration.
    model_version: str
        Model version.
    label: str, optional
        Instance label. Used for output file naming.
    site: str, optional
        Site name.
    layout_name: str, optional
        Layout name.
    array_elements: Union[str, Path, List[str]], optional
        Array element definitions (list of array element or path to file with
        the array element positions).
    """

    def __init__(
        self,
        mongo_db_config: dict,
        model_version: str,
        label: str | None = None,
        site: str | None = None,
        layout_name: str | None = None,
        array_elements: str | Path | list[str] | None = None,
    ):
        """Initialize ArrayModel."""
        self._logger = logging.getLogger(__name__)
        self._logger.debug("Init ArrayModel")
        self.mongo_db_config = mongo_db_config
        self.model_version = model_version
        self.label = label
        self.layout_name = layout_name
        self._config_file_path = None
        self._config_file_directory = None
        self.io_handler = io_handler.IOHandler()
        self.db = db_handler.DatabaseHandler(mongo_db_config=mongo_db_config)

        self.array_elements, self.site_model, self.telescope_model = self._initialize(
            site, array_elements
        )

        self._telescope_model_files_exported = False
        self._array_model_file_exported = False

    def _initialize(self, site: str, array_elements_config: str | Path | list[str]):
        """
        Initialize ArrayModel taking different configuration options into account.

        Parameters
        ----------
        site: str
            Site name.
        array_elements_config: Union[str, Path, List[str]]
            Array element definitions.

        Returns
        -------
        dict
            Dict with telescope positions.
        SiteModel
            Site model.
        dict
            Dict with telescope models.
        """
        self._logger.debug(f"Getting site parameters from DB ({site})")
        site_model = SiteModel(
            site=names.validate_site_name(site),
            mongo_db_config=self.mongo_db_config,
            model_version=self.model_version,
            label=self.label,
        )

        array_elements = {}
        # Case 1: array_elements is a file name
        if isinstance(array_elements_config, str | Path):
            array_elements = self._load_array_element_positions_from_file(
                array_elements_config, site
            )
        # Case 2: array elements is a list of elements
        elif isinstance(array_elements_config, list):
            array_elements = self._get_array_elements_from_list(array_elements_config)
        # Case 3: array elements defined in DB by array layout name
        elif self.layout_name is not None:
            array_elements = self._get_array_elements_from_list(
                site_model.get_array_elements_for_layout(self.layout_name)
            )
        if not array_elements:
            raise ValueError(
                "No array elements found. "
                "Possibly missing valid layout name or missing telescope list."
            )
        telescope_model = self._build_telescope_models(site_model, array_elements)
        return array_elements, site_model, telescope_model

    @property
    def number_of_telescopes(self) -> int:
        """
        Return the number of telescopes.

        Returns
        -------
        int
            Number of telescopes.
        """
        return len(self.telescope_model)

    @property
    def site(self) -> str:
        """
        Return site.

        Returns
        -------
        str
            Site name.
        """
        return self.site_model.site

    def _build_telescope_models(self, site_model: SiteModel, array_elements: dict) -> dict:
        """
        Build the the telescope models for all telescopes of this array.

        Includes reading of telescope model parameters from the DB.
        The array is defined in the telescopes dictionary. Array element positions
        are read from the database if no values are given in this dictionary.

        Parameters
        ----------
        site_model: SiteModel
            Site model.
        array_elements: dict
            Dict with array elements.

        Returns
        -------
        dict
            Dictionary with telescope models.
        """
        telescope_model = {}
        for element_name, _ in array_elements.items():
            collection = names.get_collection_name_from_array_element_name(element_name)
            if collection == "telescopes":
                telescope_model[element_name] = TelescopeModel(
                    site=site_model.site,
                    telescope_name=element_name,
                    model_version=self.model_version,
                    mongo_db_config=self.mongo_db_config,
                    label=self.label,
                )
        return telescope_model

    def print_telescope_list(self):
        """Print list of telescopes."""
        for tel_name, data in self.telescope_model.items():
            print(f"Name: {tel_name}\t Model: {data.name}")

    def export_simtel_telescope_config_files(self):
        """Export sim_telarray configuration files for all telescopes into the model directory."""
        exported_models = []
        for _, tel_model in self.telescope_model.items():
            name = tel_model.name + (
                "_" + tel_model.extra_label if tel_model.extra_label != "" else ""
            )
            if name not in exported_models:
                self._logger.debug(f"Exporting configuration file for telescope {name}")
                tel_model.export_config_file()
                exported_models.append(name)
            else:
                self._logger.debug(
                    f"Configuration file for telescope {name} already exists - skipping"
                )

        self._telescope_model_files_exported = True

    def export_simtel_array_config_file(self):
        """Export sim_telarray configuration file for the array into the model directory."""
        # Setting file name and the location
        config_file_name = names.simtel_config_file_name(
            array_name=self.layout_name,
            site=self.site_model.site,
            model_version=self.model_version,
            label=self.label,
        )
        self._config_file_path = self.get_config_directory().joinpath(config_file_name)

        # Writing parameters to the file
        self._logger.info(f"Writing array configuration file into {self._config_file_path}")
        simtel_writer = SimtelConfigWriter(
            site=self.site_model.site,
            layout_name=self.layout_name,
            model_version=self.model_version,
            label=self.label,
        )
        simtel_writer.write_array_config_file(
            config_file_path=self._config_file_path,
            telescope_model=self.telescope_model,
            site_model=self.site_model,
        )
        self._array_model_file_exported = True

    def export_all_simtel_config_files(self):
        """
        Export sim_telarray config file for the array and for each individual telescope.

        Config files are exported into the output model directory.
        """
        if not self._telescope_model_files_exported:
            self.export_simtel_telescope_config_files()
        if not self._array_model_file_exported:
            self.export_simtel_array_config_file()

    def get_config_file(self) -> Path:
        """
        Return the path of the array config file for sim_telarray.

        A new config file is produced if the file is not updated.

        Returns
        -------
        Path
            Path of the exported config file for sim_telarray.
        """
        self.export_all_simtel_config_files()
        return self._config_file_path

    def get_config_directory(self) -> Path:
        """
        Get the path of the array config directory for sim_telarray.

        Returns
        -------
        Path
            Path of the config directory path for sim_telarray.
        """
        if self._config_file_directory is None:
            self._config_file_directory = self.io_handler.get_output_directory(self.label, "model")
        return self._config_file_directory

    def _load_array_element_positions_from_file(
        self, array_elements_file: str | Path, site: str
    ) -> dict:
        """
        Load array element (e.g. telescope) positions from a file into a dict.

        Dictionary format: {telescope_name: {position_x: x, position_y: y, position_z: z}}

        Parameters
        ----------
        array_elements_file: Union[str, Path]
            Path to the file with the array element positions.
        site: str
            Site name.

        Returns
        -------
        dict
            Dict with telescope positions.
        """
        table = data_reader.read_table_from_file(file_name=array_elements_file)

        return {
            row["telescope_name"]: self._get_telescope_position_parameter(
                row["telescope_name"], site, row["position_x"], row["position_y"], row["position_z"]
            )
            for row in table
        }

    def _get_telescope_position_parameter(
        self,
        telescope_name: str,
        site: str,
        x: u.Quantity,
        y: u.Quantity,
        z: u.Quantity,
        parameter_version: str | None = None,
    ) -> dict:
        """
        Return dictionary with telescope position parameters (following DB model database format).

        Parameters
        ----------
        telescope_name: str
            Name of the telescope.
        site: str
            Site name.
        x: astropy.Quantity
            X ground position.
        y: astropy.Quantity
            Y ground position.
        z: astropy.Quantity
            Z ground position.

        Returns
        -------
        dict
            Dict with telescope position parameters.
        """
        return {
            "schema_version": schema.get_model_parameter_schema_version(),
            "parameter": "array_element_position_ground",
            "instrument": telescope_name,
            "site": site,
            "parameter_version": parameter_version,
            "unique_id": None,
            "value": general.convert_list_to_string(
                [x.to("m").value, y.to("m").value, z.to("m").value]
            ),
            "unit": "m",
            "type": "float64",
            "file": False,
        }

    def _get_array_elements_from_list(self, array_elements_list: list[str]) -> dict:
        """
        Return dictionary with array elements from a list of telescope names.

        Input list can contain telescope names (e.g, LSTN-01) or a telescope
        type (e.g., MSTN). In the latter case, all telescopes of this specific
        type are added.

        Parameters
        ----------
        array_elements_list: list
            List of telescope names.

        Returns
        -------
        dict
            Dict with array elements.
        """
        array_elements_dict = {}
        for name in array_elements_list:
            try:
                array_elements_dict[names.validate_array_element_name(name)] = None
            except ValueError:
                array_elements_dict.update(self._get_all_array_elements_of_type(name))
        return array_elements_dict

    def _get_all_array_elements_of_type(self, array_element_type: str) -> dict:
        """
        Return all array elements of a specific type using the database.

        Parameters
        ----------
        array_element_type : str
            Type of the array element (e.g. LSTN, MSTS)

        Returns
        -------
        dict
            Dict with array elements.
        """
        all_elements = self.db.get_array_elements_of_type(
            array_element_type=array_element_type,
            model_version=self.model_version,
            collection="telescopes",
        )
        return self._get_array_elements_from_list(all_elements)

    def export_array_elements_as_table(self, coordinate_system: str = "ground") -> QTable:
        """
        Export array elements positions to astropy table.

        Parameters
        ----------
        coordinate_system: str
            Positions are exported in this coordinate system.

        Returns
        -------
        astropy.table.QTable
            Astropy table with the telescope layout information.
        """
        table = QTable(meta={"array_name": self.layout_name, "site": self.site_model.site})

        name, pos_x, pos_y, pos_z, tel_r = [], [], [], [], []
        for tel_name, data in self.telescope_model.items():
            name.append(tel_name)
            xyz = data.position(coordinate_system=coordinate_system)
            pos_x.append(xyz[0])
            pos_y.append(xyz[1])
            pos_z.append(xyz[2])
            try:
                # add tests of KeyError after positions calibration_elements are added to DB
                tel_r.append(data.get_parameter_value_with_unit("telescope_sphere_radius"))
            except KeyError:  # not all array elements have a sphere radius
                tel_r.append(0.0 * u.m)

        table["telescope_name"] = name
        if coordinate_system == "ground":
            table["position_x"] = pos_x
            table["position_y"] = pos_y
            table["position_z"] = pos_z
        elif coordinate_system == "utm":
            table["utm_east"] = pos_x
            table["utm_north"] = pos_y
            table["altitude"] = pos_z
        table["sphere_radius"] = tel_r

        table.sort("telescope_name")
        return table



#!/usr/bin/python3

r"""
Derives the limits for energy, radial distance, and viewcone to be used in CORSIKA simulations.

The limits are derived based on the event loss fraction specified by the user.

Command line arguments
----------------------
event_data_file (str, required)
    Path to the file containing the event data.
loss_fraction (float, required)
    Fraction of events to be lost.
telescope_ids (list of int, optional)
    List of telescope IDs to filter the events (default is None).
    Definition of the telescope IDs can be found
    in the telescope definition file used for simulations.


Example
-------
Derive limits for a given file with a specified loss fraction.

.. code-block:: console

    simtools-production-derive-corsika-limits\\
        --event_data_file path/to/event_data_file.hdf5 \\
        --loss_fraction 1e-6 \\
        --telescope_ids 1 2 3
"""

import logging

import simtools.utils.general as gen
from simtools.configuration import configurator
from simtools.production_configuration.derive_corsika_limits import LimitCalculator

_logger = logging.getLogger(__name__)


def _parse():
    """
    Parse command line configuration.

    Parameters
    ----------
    event_data_file: str
        The event data file.
    loss_fraction: float
        Loss fraction of events for limit derivation.
    telescope_ids: list of int, optional
        List of telescope IDs to filter the events (default is None).
        Definition of the telescope IDs can be found in the telescope
        definition file used for simulations.

    Returns
    -------
    CommandLineParser
        Command line parser object

    """
    config = configurator.Configurator(
        description="Derive limits for energy, radial distance, and viewcone."
    )
    config.parser.add_argument(
        "--event_data_file",
        type=str,
        required=True,
        help="Path to the event data file containing the event data.",
    )
    config.parser.add_argument(
        "--loss_fraction", type=float, required=True, help="Fraction of events to be lost."
    )
    config.parser.add_argument(
        "--telescope_ids",
        type=int,
        nargs="*",
        help="List of telescope IDs to filter the events as used in sim_telarray.",
    )
    return config.initialize(db_config=False)


def main():
    """Derive limits for energy, radial distance, and viewcone."""
    args_dict, _ = _parse()

    logger = logging.getLogger()
    logger.setLevel(gen.get_log_level_from_user(args_dict["log_level"]))

    event_data_file_path = args_dict["event_data_file"]
    loss_fraction = args_dict["loss_fraction"]
    telescope_ids = args_dict.get("telescope_ids")

    calculator = LimitCalculator(event_data_file_path, telescope_list=telescope_ids)

    lower_energy_limit = calculator.compute_lower_energy_limit(loss_fraction)
    _logger.info(f"Lower energy threshold: {lower_energy_limit}")

    upper_radial_distance = calculator.compute_upper_radial_distance(loss_fraction)
    _logger.info(f"Upper radius threshold: {upper_radial_distance}")

    viewcone = calculator.compute_viewcone(loss_fraction)
    _logger.info(f"Viewcone radius: {viewcone}")


if __name__ == "__main__":
    main()



"""Derives simulation configuration parameters for a grid point based on several metrics."""

from simtools.production_configuration.calculate_statistical_errors_grid_point import (
    StatisticalErrorEvaluator,
)
from simtools.production_configuration.event_scaler import EventScaler

__all__ = ["SimulationConfig"]


class SimulationConfig:
    """
    Configures simulation parameters for a specific grid point.

    Parameters
    ----------
    grid_point : dict
        Dictionary representing a grid point with azimuth, elevation, and night sky background.
    file_path : str
        Path to the DL2 MC event file for statistical uncertainty evaluation.
    file_type : str
        Type of the DL2 MC event file ('point-like' or 'cone').
    metrics : dict, optional
        Dictionary of metrics to evaluate.
    """

    def __init__(
        self,
        grid_point: dict[str, float],
        file_path: str,
        file_type: str,
        metrics: dict[str, float] | None = None,
    ):
        """Initialize the simulation configuration for a grid point."""
        self.grid_point = grid_point
        self.file_path = file_path
        self.file_type = file_type
        self.metrics = metrics or {}
        self.evaluator = StatisticalErrorEvaluator(file_path, file_type, metrics)
        self.event_scaler = EventScaler(self.evaluator, self.metrics)
        self.simulation_params = {}

    def configure_simulation(self) -> dict[str, float]:
        """
        Configure the simulation parameters for the grid point.

        Returns
        -------
        dict
            A dictionary with simulation parameters such as core scatter area,
              viewcone, and number of simulated events.
        """
        self.simulation_params = {
            "core_scatter_area": self._calculate_core_scatter_area(),
            "viewcone": self._calculate_viewcone(),
            "number_of_events": self.calculate_required_events(),
        }
        return self.simulation_params

    def calculate_required_events(self) -> int:
        """
        Calculate the required number of simulated events based on statistical error metrics.

        Uses the EventScaler to scale the events.

        Returns
        -------
        int
            The number of simulated events required.
        """
        return self.event_scaler.scale_events()

    def _calculate_core_scatter_area(self) -> float:
        """
        Calculate the core scatter area based on the grid point.

        Returns
        -------
        float
            The core scatter area.
        """
        base_area = self._fetch_simulated_core_scatter_area()
        area_factor = self._get_area_factor_from_grid_point()
        return base_area * area_factor

    def _calculate_viewcone(self) -> float:
        """
        Calculate the viewcone based on the grid point conditions.

        Returns
        -------
        float
            The viewcone value.
        """
        base_viewcone = self._fetch_simulated_viewcone()
        viewcone_factor = self._get_viewcone_factor_from_grid_point()

        return base_viewcone * viewcone_factor

    def _get_area_factor_from_grid_point(self) -> float:
        """
        Determine the area factor.

        The area factor is based on the grid point's azimuth,
          elevation, and night sky background.

        Returns
        -------
        float
            The area factor.
        """
        azimuth = self.grid_point.get("azimuth", 0)
        elevation = self.grid_point.get("elevation", 0)
        night_sky_background = self.grid_point.get("night_sky_background", 0)

        # Implement reading of factor from LUT
        return azimuth + elevation + night_sky_background

    def _get_viewcone_factor_from_grid_point(self) -> float:
        """
        Determine the viewcone factor.

        The factor is based on the grid point's azimuth,
          elevation, and night sky background.

        Returns
        -------
        float
            The viewcone factor.
        """
        azimuth = self.grid_point.get("azimuth", 0)
        elevation = self.grid_point.get("elevation", 0)
        night_sky_background = self.grid_point.get("night_sky_background", 0)

        # Implement reading of factor from LUT
        return azimuth + elevation + night_sky_background

    def _fetch_simulated_core_scatter_area(self) -> float:
        """
        Fetch the core scatter area from existing simulated files based on grid point conditions.

        Returns
        -------
        float
            The fetched core scatter outer bound.
        """
        return self.evaluator.data["core_range"]

    def _fetch_simulated_viewcone(self) -> float:
        """
        Fetch the viewcone from existing simulated files based on grid point conditions.

        Returns
        -------
        float
            The fetched viewcone outer bound.
        """
        return self.evaluator.data["viewcone"]



#!/usr/bin/python3

r"""
    Simulate calibration devices using the light emission package.

    Run the application in the command line.
    There are two ways this application can be executed:

    1. Illuminator at varying distances.
    2. Illuminator and telescopes at fixed positions as defined in the layout.

    Example Usage
    -------------

    1. Simulate light emission with varying distances:

        .. code-block:: console

            simtools-simulate-light-emission --telescope MSTN-04 --site North \
            --illuminator ILLN-01 --light_source_setup variable \
            --model_version 6.0.0 --light_source_type led

    2. Simulate light emission with telescopes at fixed positions according to the layout:

        .. code-block:: console

            simtools-simulate-light-emission --telescope MSTN-04 --site North \
            --illuminator ILLN-01 --light_source_setup layout \
            --model_version 6.0.0 \
            --light_source_type led

    Command Line Arguments
    ----------------------

    --telescope (str, required)
        Telescope model name (e.g. LSTN-01, SSTS-design, SSTS-25, ...)

    --site (str, required)
        Site name (North or South).

    --illuminator (str, optional)
        Illuminator in array, e.g., ILLN-01.

    --light_source_setup (str, optional)
        Select calibration light source positioning/setup:
        - "variable" for varying distances.
        - "layout" for actual telescope positions.

    --model_version (str, optional)
        Version of the simulation model.

    --light_source_type (str, optional)
        Select calibration light source type: led (default) or laser.
        This changes the pre-compiled (simtel_array) application that is used to run the
        light emission package with. Currently we use xyzls (laser), and ls-beam can be
        accessed by using the laser option.

    --off_axis_angle (float, optional)
        Off axis angle for light source direction.

    --plot (flag, optional)
        Produce a multiple pages pdf file with the image plots.


    Example
    -------

    Simulate isotropic light source at different distances for the MSTN-04:

    .. code-block:: console

        simtools-simulate-light-emission --telescope MSTN-04 --site North \
        --illuminator ILLN-01 --light_source_setup variable \
        --model_version 6.0.0 --light_source_type led    ```

    Expected Output:

    .. code-block:: console

        light-emission package stage:
        File '/workdir/external/simtools/simtools-output/light_emission/model/
            atmprof_ecmwf_north_winter_fixed.dat' registered for atmospheric profile 99.
        Atmospheric profile 99 to be read from file '/workdir/external/simtools/
            simtools-output/light_emission/model/atmprof_ecmwf_north_winter_fixed.dat'.
        Atmospheric profile 99 with 55 levels read from file /workdir/external/
            simtools/simtools-output/light_emission/model/atmprof_ecmwf_north_winter_fixed.dat
        Initialize atmosphere ranging from 0.000 to 120.000 km a.s.l.
        IACT control parameter line: print_events 999 10 100 1000 0
        Case 1: 1 event with 1e+10 photons.
        Using IACT/ATMO package version 1.67 (2023-11-10) for CORSIKA 6.999
        Output file /workdir/external/simtools/simtools-output/light_emission/xyzls.iact.gz
            not yet created.
        Telescope output file: '/workdir/external/simtools/simtools-output/
            light_emission/xyzls.iact.gz'
        ....
        ....
        Sim_telarray stage:
        Telescope 1 triggered (1/0/0/0, mask 1), summation from 36 to 95 of 105
        Event end data has been found.
        Shower of 0.0000 TeV energy was seen in 1 of 1 cases.
        Photon statistics:
        All photons:               928518
        Used photons:              928518
        Not absorbed/max. Q.E.:    189560
        Reflected on mirror:        26815
        Camera hit:                 25574
        Pixel hit:                  25574
        Detected:                   20998
        Trigger statistics:
        Tel. triggered:             1
        Tel. + array:               1
        Early readout:              0
        Late readout:               0
        Finish data conversion ...
        Writing 13 histograms to output file.


"""

import logging
import subprocess
from pathlib import Path

import astropy.units as u

import simtools.utils.general as gen
from simtools.configuration import configurator
from simtools.corsika.corsika_histograms_visualize import save_figs_to_pdf
from simtools.model.calibration_model import CalibrationModel
from simtools.model.site_model import SiteModel
from simtools.model.telescope_model import TelescopeModel
from simtools.simtel.simulator_light_emission import SimulatorLightEmission
from simtools.visualization.visualize import plot_simtel_ctapipe


def _parse(label):
    """Parse command line configuration."""
    config = configurator.Configurator(
        label=label,
        description=(
            "Simulate the light emission by an artificial light source for calibration purposes."
        ),
    )
    config.parser.add_argument(
        "--off_axis_angle",
        help="Off axis angle for light source direction",
        type=float,
        default=0.0,
        required=False,
    )
    config.parser.add_argument(
        "--plot",
        help="Produce a multiple pages pdf file with the image plots.",
        action="store_true",
    )
    config.parser.add_argument(
        "--light_source_type",
        help="Select calibration light source type: led or laser",
        type=str,
        default="led",
        choices=["led", "laser"],
        required=False,
    )
    config.parser.add_argument(
        "--light_source_setup",
        help="Select calibration light source positioning/setup: \
              varying distances (variable), layout positions (layout)",
        type=str,
        choices=["layout", "variable"],
        default=None,
    )
    config.parser.add_argument(
        "--distances_ls",
        help="Light source distance in m (Example --distances_ls 800 1200)",
        nargs="+",
        required=False,
    )
    config.parser.add_argument(
        "--illuminator",
        help="Illuminator in array, i.e. ILLN-design",
        type=str,
        default=None,
    )
    config.parser.add_argument(
        "--telescope_file",
        help="Telescope position file (temporary)",
        type=str,
        default=None,
    )
    config.parser.add_argument(
        "--return_cleaned",
        help="ctapipe, if image should be cleaned, \
              notice as well image cleaning parameters",
        type=str,
        default=False,
        required=False,
    )
    config.parser.add_argument(
        "--picture_thresh",
        help="ctapipe, threshold above which all pixels are retained",
        type=int,
        required=False,
    )
    config.parser.add_argument(
        "--boundary_thresh",
        help="ctapipe, threshold above which pixels are retained if\
              they have a neighbor already above the picture_thresh",
        type=int,
        required=False,
    )
    config.parser.add_argument(
        "--min_neighbors",
        help="ctapipe, A picture pixel survives cleaning only if it has at\
              least this number of picture neighbors. This has no effect in\
              case keep_isolated_pixels is True",
        type=int,
        required=False,
    )
    config.parser.add_argument(
        "--level",
        help="read 5",
        type=int,
        default=5,
        required=False,
    )
    config.parser.add_argument(
        "--integration_window",
        help="ctapipe, A picture pixel survives cleaning only if it has at\
              least this number of picture neighbors. This has no effect in\
              case keep_isolated_pixels is True",
        nargs="*",
        default=["7", "3"],
        required=False,
    )
    return config.initialize(
        db_config=True,
        simulation_model=["telescope", "model_version"],
        require_command_line=True,
    )


def distance_list(arg):
    """
    Convert distance list to astropy quantities.

    Parameters
    ----------
    arg: list
        List of distances.

    Returns
    -------
    values: list
        List of distances as astropy quantities.
    """
    try:
        return [float(x) * u.m for x in arg]
    except ValueError as exc:
        raise ValueError("Distances must be numeric values") from exc


def default_le_configs(le_application, args_dict):
    """
    Define default light emission configurations.

    Predefined angular distribution names not requiring to read any table are
    "Isotropic", "Gauss:<rms>", "Rayleigh", "Cone:<angle>", and "FilledCone:<angle>", "Parallel",
    with all angles given in degrees, all with respect to the given direction vector
    (vertically downwards if missing). If the light source has a non-zero length and velocity
    (in units of the vacuum speed of light), it is handled as a moving source,
    in the given direction.

    Parameters
    ----------
    le_application: str
        Light emission application.

    args_dict: dict
        Dictionary with command line arguments.

    Returns
    -------
    default_config: dict
        Default light emission configuration.
    """
    if le_application in ("xyzls", "ls-beam") and args_dict["light_source_setup"] == "variable":
        return {
            "x_pos": {
                "len": 1,
                "unit": u.Unit("cm"),
                "default": 0 * u.cm,
                "names": ["x_position"],
            },
            "y_pos": {
                "len": 1,
                "unit": u.Unit("cm"),
                "default": 0 * u.cm,
                "names": ["y_position"],
            },
            "z_pos": {
                "len": 1,
                "unit": u.Unit("cm"),
                "default": [i * 100 for i in [200, 300, 400, 600, 800, 1200, 2000, 4000]] * u.cm,
                "names": ["z_position"],
            },
            "direction": {
                "len": 3,
                "unit": u.dimensionless_unscaled,
                "default": [0, 0, -1],
                "names": ["direction", "cx,cy,cz"],
            },
        }

    return {}


def select_application(args_dict):
    """
    Select sim_telarray application for light emission simulations.

    Parameters
    ----------
    args_dict: dict
        Dictionary with command line arguments.

    Returns
    -------
    le_application: str
        Light emission application.
    """
    le_application = None
    if args_dict["light_source_type"] == "led":
        le_application = "xyzls"
    elif args_dict["light_source_type"] == "laser":
        le_application = "ls-beam"

    return le_application, args_dict["light_source_setup"]


def main():
    """Simulate light emission."""
    label = Path(__file__).stem
    args_dict, db_config = _parse(label)
    le_application = select_application(args_dict)
    default_le_config = default_le_configs(le_application[0], args_dict)
    logger = logging.getLogger()
    logger.setLevel(gen.get_log_level_from_user(args_dict["log_level"]))

    telescope_model = TelescopeModel(
        site=args_dict["site"],
        telescope_name=args_dict["telescope"],
        mongo_db_config=db_config,
        model_version=args_dict["model_version"],
        label=label,
    )

    calibration_model = CalibrationModel(
        site=args_dict["site"],
        calibration_device_model_name=args_dict["illuminator"],
        mongo_db_config=db_config,
        model_version=args_dict["model_version"],
        label=label,
    )

    site_model = SiteModel(
        site=args_dict["site"],
        mongo_db_config=db_config,
        model_version=args_dict["model_version"],
        label=label,
    )

    if args_dict["light_source_setup"] == "variable":
        if args_dict["distances_ls"] is not None:
            default_le_config["z_pos"]["default"] = distance_list(args_dict["distances_ls"])
        logger.info(f"Simulating for distances of {default_le_config['z_pos']['default']}")
        figures = []
        for distance in default_le_config["z_pos"]["default"]:
            le_config = default_le_config.copy()
            le_config["z_pos"]["default"] = distance
            light_source = SimulatorLightEmission(
                telescope_model=telescope_model,
                calibration_model=calibration_model,
                site_model=site_model,
                default_le_config=le_config,
                le_application=le_application,
                simtel_path=args_dict["simtel_path"],
                light_source_type=args_dict["light_source_type"],
            )
            run_script = light_source.prepare_script(generate_postscript=True, **args_dict)
            log_file = f"{light_source.output_directory}/logfile.log"
            with open(log_file, "w", encoding="utf-8") as log_file:
                subprocess.run(
                    run_script,
                    shell=False,
                    check=False,
                    text=True,
                    stdout=log_file,
                    stderr=log_file,
                )

                try:
                    filename = (
                        f"{light_source.output_directory}/"
                        f"{light_source.le_application[0]}_"
                        f"{light_source.le_application[1]}.simtel.gz"
                    )
                    fig = plot_simtel_ctapipe(
                        filename,
                        cleaning_args=[
                            args_dict["boundary_thresh"],
                            args_dict["picture_thresh"],
                            args_dict["min_neighbors"],
                        ],
                        distance=light_source.default_le_config["z_pos"]["default"],
                        return_cleaned=args_dict["return_cleaned"],
                    )
                    figures.append(fig)
                except AttributeError:
                    msg = "telescope not triggered at distance of"
                    msg += f"{light_source.distance.to(u.meter)}"
                    logger.warning(msg)

                save_figs_to_pdf(
                    figures,
                    f"{light_source.output_directory}/{args_dict['telescope']}_"
                    f"{light_source.le_application[0]}_"
                    f"{light_source.le_application[1]}.pdf",
                )

    elif args_dict["light_source_setup"] == "layout":
        light_source = SimulatorLightEmission(
            telescope_model=telescope_model,
            calibration_model=calibration_model,
            site_model=site_model,
            default_le_config=default_le_config,
            le_application=le_application,
            simtel_path=args_dict["simtel_path"],
            light_source_type=args_dict["light_source_type"],
        )
        run_script = light_source.prepare_script(generate_postscript=True, **args_dict)
        log_file = f"{light_source.output_directory}/logfile.log"
        with open(log_file, "w", encoding="utf-8") as log_file:
            subprocess.run(
                run_script, shell=False, check=False, text=True, stdout=log_file, stderr=log_file
            )
            try:
                filename = (
                    f"{light_source.output_directory}/"
                    f"{light_source.le_application[0]}_{light_source.le_application[1]}.simtel.gz"
                )
                fig = plot_simtel_ctapipe(
                    filename,
                    cleaning_args=[
                        args_dict["boundary_thresh"],
                        args_dict["picture_thresh"],
                        args_dict["min_neighbors"],
                    ],
                    distance=light_source.distance,
                    return_cleaned=args_dict["return_cleaned"],
                )
            except AttributeError:
                msg = f"telescope not triggered at distance of {light_source.distance.to(u.meter)}"
                logger.warning(msg)

        save_figs_to_pdf(
            [fig],
            f"{light_source.output_directory}/{args_dict['telescope']}_"
            f"{light_source.le_application[0]}_"
            f"{light_source.le_application[1]}.pdf",
        )


if __name__ == "__main__":
    main()



#!/usr/bin/python3
"""Definition of site model."""

import logging
from pathlib import Path

from simtools.model.model_parameter import ModelParameter

__all__ = ["SiteModel"]


class SiteModel(ModelParameter):
    """
    SiteModel represents the MC model of an observatory site.

    Parameters
    ----------
    site: str
        Site name (e.g., South or North).
    mongo_db_config: dict
        MongoDB configuration.
    model_version: str
        Model version.
    label: str, optional
        Instance label. Important for output file naming.
    """

    def __init__(
        self,
        site: str,
        mongo_db_config: dict,
        model_version: str,
        label: str | None = None,
    ):
        """Initialize SiteModel."""
        self._logger = logging.getLogger(__name__)
        self._logger.debug("Init SiteModel for site %s", site)
        super().__init__(
            site=site,
            mongo_db_config=mongo_db_config,
            model_version=model_version,
            db=None,
            label=label,
        )

    def get_reference_point(self) -> dict:
        """
        Get reference point coordinates as dict.

        Returns
        -------
        dict
            Reference point coordinates as dict
        """
        return {
            "center_altitude": self.get_parameter_value_with_unit("reference_point_altitude"),
            "center_northing": self.get_parameter_value_with_unit("reference_point_utm_north"),
            "center_easting": self.get_parameter_value_with_unit("reference_point_utm_east"),
            "epsg_code": self.get_parameter_value("epsg_code"),
        }

    def get_corsika_site_parameters(
        self, config_file_style: bool = False, model_directory: Path | None = None
    ) -> dict:
        """
        Get site-related CORSIKA parameters as dict.

        Parameters are returned with units wherever possible.

        Parameters
        ----------
        config_file_style: bool
            Return using CORSIKA config file syntax
        model_directory: Path, optional
            Model directory to use for file paths

        Returns
        -------
        dict
            Site-related CORSIKA parameters as dict
        """
        if config_file_style:
            model_directory = model_directory or Path()
            return {
                "OBSLEV": [
                    self.get_parameter_value_with_unit("corsika_observation_level").to_value("cm")
                ],
                # We always use a custom profile by filename, so this has to be set to 99
                "ATMOSPHERE": [99, "Y"],
                "IACT ATMOFILE": [
                    model_directory / self.get_parameter_value("atmospheric_profile")
                ],
                "MAGNET": [
                    self.get_parameter_value("geomag_horizontal"),
                    self.get_parameter_value("geomag_vertical"),
                ],
                "ARRANG": [self.get_parameter_value("geomag_rotation")],
            }

        return {
            "corsika_observation_level": self.get_parameter_value_with_unit(
                "corsika_observation_level"
            ),
            "geomag_horizontal": self.get_parameter_value_with_unit("geomag_horizontal"),
            "geomag_vertical": self.get_parameter_value_with_unit("geomag_vertical"),
            "geomag_rotation": self.get_parameter_value_with_unit("geomag_rotation"),
        }

    def get_array_elements_for_layout(self, layout_name: str) -> list:
        """
        Return list of array elements for a given array layout.

        Parameters
        ----------
        layout_name: str
            Name of the array layout

        Returns
        -------
        list
            List of array elements
        """
        layouts = self.get_parameter_value("array_layouts")
        for layout in layouts:
            if layout["name"] == layout_name.lower():
                return layout["elements"]
        raise ValueError(f"Array layout '{layout_name}' not found in '{self.site}' site model.")

    def get_list_of_array_layouts(self) -> list:
        """
        Get list of available array layouts.

        Returns
        -------
        list
            List of available array layouts
        """
        return [layout["name"] for layout in self.get_parameter_value("array_layouts")]

    def export_atmospheric_transmission_file(self, model_directory: Path):
        """
        Export atmospheric transmission file.

        This method is needed because when CORSIKA is not piped to sim_telarray,
        the atmospheric transmission file is not written out to the model directory.
        This method allows to export it explicitly.

        Parameters
        ----------
        model_directory: Path
            Model directory to export the file to.
        """
        self.db.export_model_files(
            parameters={
                "atmospheric_transmission_file": {
                    "value": self.get_parameter_value("atmospheric_profile"),
                    "file": True,
                }
            },
            dest=model_directory,
        )



#!/usr/bin/python3
"""
Generate sim_telarray configuration files for a given array.

The applications generates the sim_telarray configuration files for a given array, site,
and model_version using the model parameters stored in the database.

Command line arguments
----------------------
site : str
    Site name (e.g., North, South).
array_layout_name : str
    Name of the layout array (e.g., test_layout, alpha, 4mst, etc.).
array_element_list : list
    List of array elements (e.g., telescopes) to plot (e.g., ``LSTN-01 LSTN-02 MSTN``).

Example
-------
North - 5.0.0:

.. code-block:: console

    simtools-generate-array-config --site North --array_layout_name alpha --model_version 5.0.0

The output is saved in simtools-output/test/model.
"""

import logging
from pathlib import Path

import simtools.utils.general as gen
from simtools.configuration import configurator
from simtools.model.array_model import ArrayModel


def _parse(label, description):
    """
    Parse command line configuration.

    Parameters
    ----------
    label : str
        Label describing the application.
    description : str
        Description of the application.

    Returns
    -------
    CommandLineParser
        Command line parser object.
    """
    config = configurator.Configurator(label=label, description=description)
    return config.initialize(db_config=True, simulation_model=["site", "layout", "model_version"])


def main():
    """Generate sim_telarray configuration files for a given array."""
    args_dict, db_config = _parse(
        label=Path(__file__).stem,
        description=("Generate sim_telarray configuration files for a given array."),
    )
    logger = logging.getLogger("simtools")
    logger.setLevel(gen.get_log_level_from_user(args_dict["log_level"]))

    array_model = ArrayModel(
        label=args_dict["label"],
        model_version=args_dict["model_version"],
        mongo_db_config=db_config,
        site=args_dict.get("site"),
        layout_name=args_dict.get("array_layout_name"),
        array_elements=args_dict.get("array_elements"),
    )
    array_model.print_telescope_list()
    array_model.export_all_simtel_config_files()


if __name__ == "__main__":
    main()



"""Project wide constants."""

from importlib.resources import files

# Schema path
SCHEMA_PATH = files("simtools") / "schemas"
# Path to metadata jsonschema
METADATA_JSON_SCHEMA = SCHEMA_PATH / "metadata.metaschema.yml"
# Path to plotting configuration json schema
PLOT_CONFIG_SCHEMA = SCHEMA_PATH / "plot_configuration.metaschema.yml"
# Path to model parameter metaschema
MODEL_PARAMETER_METASCHEMA = SCHEMA_PATH / "model_parameter.metaschema.yml"
# Path to model parameter description metaschema
MODEL_PARAMETER_DESCRIPTION_METASCHEMA = (
    SCHEMA_PATH / "model_parameter_and_data_schema.metaschema.yml"
)
# Path to model parameter schema files
MODEL_PARAMETER_SCHEMA_PATH = SCHEMA_PATH / "model_parameters"
# URL to model parameter schema files
MODEL_PARAMETER_SCHEMA_URL = (
    "https://raw.githubusercontent.com/gammasim/simtools/main/src/simtools/schemas/"
    "/model_parameters"
)
